/*! \file
   This header defines utility functions to deal with complex numbers and complex 
   polynomials.*/

/*! Returns the complex conjugate of the given complex number (i.e. it changes the 
   sign of the y-component).*/
float2 Conjugate(float2 Z){
	return float2(Z.x,-Z.y);
}
/*! This function implements complex multiplication.*/
float2 Multiply(float2 LHS,float2 RHS){
	return float2(LHS.x*RHS.x-LHS.y*RHS.y,LHS.x*RHS.y+LHS.y*RHS.x);
}
/*! This function computes the magnitude of the given complex number.*/
float Magnitude(float2 Z){
	return sqrt(dot(Z,Z));
}
/*! This function computes the quotient of two complex numbers. The denominator must 
   not be zero.*/
float2 Divide(float2 Numerator,float2 Denominator){
	return float2(Numerator.x*Denominator.x+Numerator.y*Denominator.y,-Numerator.x*Denominator.y+Numerator.y*Denominator.x)/dot(Denominator,Denominator);
}
/*! This function implements computation of the reciprocal of the given non-zero 
   complex number.*/
float2 Reciprocal(float2 Z){
	return float2(Z.x,-Z.y)/dot(Z,Z);
}
/*! This utility function implements complex squaring.*/
float2 Square(float2 Z){
	return float2(Z.x*Z.x-Z.y*Z.y,2.0f*Z.x*Z.y);
}
/*! This utility function implements complex computation of the third power.*/
float2 Cube(float2 Z){
	return Multiply(Square(Z),Z);
}
/*! This utility function computes one root of the given complex value. The other 
   one can be found using the unary minus operator.
  \warning This function is continuous but not defined on the negative real axis 
		   (and cannot be continued continuously there).
  \sa SquareRoot() */
float2 SquareRootUnsafe(float2 Z){
	float ZLengthSq=dot(Z,Z);
	float ZLengthInv=rsqrt(ZLengthSq);
	float2 UnnormalizedRoot=Z*ZLengthInv+float2(1.0f,0.0f);
	float UnnormalizedRootLengthSq=dot(UnnormalizedRoot,UnnormalizedRoot);
	float NormalizationFactorInvSq=UnnormalizedRootLengthSq*ZLengthInv;
	float NormalizationFactor=rsqrt(NormalizationFactorInvSq);
	return NormalizationFactor*UnnormalizedRoot;
}
/*! This utility function computes one square root of the given complex value. The 
   other one can be found using the unary minus operator.
  \note This function has discontinuities for values with real part zero.
  \sa SquareRootUnsafe() */
float2 SquareRoot(float2 Z){
	float2 ZPositiveRealPart=float2(abs(Z.x),Z.y);
	float2 ComputedRoot=SquareRootUnsafe(ZPositiveRealPart);
	return (Z.x>=0.0)?ComputedRoot:ComputedRoot.yx;
}
/*! This utility function computes one cubic root of the given complex value. The 
   other roots can be found by multiplication by cubic roots of unity.
  \note This function has various discontinuities.*/
float2 CubicRoot(float2 Z){
	float Argument=atan2(Z.y,Z.x);
	float NewArgument=Argument/3.0f;
	float2 NormalizedRoot;
	sincos(NewArgument,NormalizedRoot.y,NormalizedRoot.x);
	return NormalizedRoot*pow(dot(Z,Z),1.0f/6.0f);
}

/*! @{
   Returns the complex conjugate of the given complex vector (i.e. it changes the 
   second column resp the y-component).*/
float2x2 Conjugate(float2x2 Vector){
	return float2x2(Vector[0].x,-Vector[0].y,Vector[1].x,-Vector[1].y);
}
float3x2 Conjugate(float3x2 Vector){
	return float3x2(Vector[0].x,-Vector[0].y,Vector[1].x,-Vector[1].y,Vector[2].x,-Vector[2].y);
}
float4x2 Conjugate(float4x2 Vector){
	return float4x2(Vector[0].x,-Vector[0].y,Vector[1].x,-Vector[1].y,Vector[2].x,-Vector[2].y,Vector[3].x,-Vector[3].y);
}
void Conjugate(out float2 OutConjugateVector[5],float2 Vector[5]){
	[unroll] for(int i=0;i!=5;++i){
		OutConjugateVector[i]=float2(Vector[i].x,-Vector[i].x);
	}
}
//!@}


#define POLYNOMIAL_DEGREE 1
/*! This function evaluates the polynomial with the given coefficients at the given 
   value. The complex number given by pCoefficient[j] is used as coefficient for the 
   j-th power. All other coefficients are zero.*/
float2 EvaluatePolynomial(float2 pCoefficient[POLYNOMIAL_DEGREE+1],float2 X){
	float2 Result=pCoefficient[POLYNOMIAL_DEGREE];
	[unroll] for(int i=POLYNOMIAL_DEGREE-1;i>=0;--i){
		Result=Multiply(X,Result)+pCoefficient[i];
	}
	return Result;
}
#define POLYNOMIAL_DEGREE 2
float2 EvaluatePolynomial(float2 pCoefficient[POLYNOMIAL_DEGREE+1],float2 X){
	float2 Result=pCoefficient[POLYNOMIAL_DEGREE];
	[unroll] for(int i=POLYNOMIAL_DEGREE-1;i>=0;--i){
		Result=Multiply(X,Result)+pCoefficient[i];
	}
	return Result;
}
#define POLYNOMIAL_DEGREE 3
float2 EvaluatePolynomial(float2 pCoefficient[POLYNOMIAL_DEGREE+1],float2 X){
	float2 Result=pCoefficient[POLYNOMIAL_DEGREE];
	[unroll] for(int i=POLYNOMIAL_DEGREE-1;i>=0;--i){
		Result=Multiply(X,Result)+pCoefficient[i];
	}
	return Result;
}
#define POLYNOMIAL_DEGREE 4
float2 EvaluatePolynomial(float2 pCoefficient[POLYNOMIAL_DEGREE+1],float2 X){
	float2 Result=pCoefficient[POLYNOMIAL_DEGREE];
	[unroll] for(int i=POLYNOMIAL_DEGREE-1;i>=0;--i){
		Result=Multiply(X,Result)+pCoefficient[i];
	}
	return Result;
}
#define POLYNOMIAL_DEGREE 5
float2 EvaluatePolynomial(float2 pCoefficient[POLYNOMIAL_DEGREE+1],float2 X){
	float2 Result=pCoefficient[POLYNOMIAL_DEGREE];
	[unroll] for(int i=POLYNOMIAL_DEGREE-1;i>=0;--i){
		Result=Multiply(X,Result)+pCoefficient[i];
	}
	return Result;
}
#define POLYNOMIAL_DEGREE 6
float2 EvaluatePolynomial(float2 pCoefficient[POLYNOMIAL_DEGREE+1],float2 X){
	float2 Result=pCoefficient[POLYNOMIAL_DEGREE];
	[unroll] for(int i=POLYNOMIAL_DEGREE-1;i>=0;--i){
		Result=Multiply(X,Result)+pCoefficient[i];
	}
	return Result;
}
#define POLYNOMIAL_DEGREE 7
float2 EvaluatePolynomial(float2 pCoefficient[POLYNOMIAL_DEGREE+1],float2 X){
	float2 Result=pCoefficient[POLYNOMIAL_DEGREE];
	[unroll] for(int i=POLYNOMIAL_DEGREE-1;i>=0;--i){
		Result=Multiply(X,Result)+pCoefficient[i];
	}
	return Result;
}
#define POLYNOMIAL_DEGREE 8
float2 EvaluatePolynomial(float2 pCoefficient[POLYNOMIAL_DEGREE+1],float2 X){
	float2 Result=pCoefficient[POLYNOMIAL_DEGREE];
	[unroll] for(int i=POLYNOMIAL_DEGREE-1;i>=0;--i){
		Result=Multiply(X,Result)+pCoefficient[i];
	}
	return Result;
}
