/*! \file
   This header defines utility functions that help to avoid code duplication among 
   different techniques for integrating single scattering in participating media.*/

/*! Converts the given rectified coordinates to Cartesian coordinates in 
   rectification space and returns the result.
  \sa CartesianToRectified() */
float3 RectifiedToCartesian(float Distance,float Inclination,float Azimuth){
	float AzimuthSine,AzimuthCosine;
	sincos(Azimuth,AzimuthSine,AzimuthCosine);
	float InclinationSine,InclinationCosine;
	sincos(Inclination,InclinationSine,InclinationCosine);
	return Distance*float3(AzimuthCosine,AzimuthSine,InclinationCosine/InclinationSine);
}


/*! This function converts the given position in Cartesian coordinates of 
   rectification space into rectified coordinates. It is a right inverse of 
   RectifiedToCartesian().
  \param OutDistance The non-negative distance from the origin after projecting to 
		 the z=0 plane.
  \param OutInclination The angle to the z-axis in radians from zero to pi.
  \param OutAzimuth The planar angle to the x-axis in radians from zero to two pi.
  \param CartesianCoordinates The Cartesian coordinates that are to be converted.*/
void CartesianToRectified(out float OutDistance,out float OutInclination,out float OutAzimuth,float3 CartesianCoordinates){
	const float PI=3.1415926535897932384626433832795f;
	OutDistance=length(CartesianCoordinates.xy);
	OutInclination=atan2(OutDistance,CartesianCoordinates.z);
	OutAzimuth=atan2(CartesianCoordinates.y,CartesianCoordinates.x);
	OutAzimuth=(OutAzimuth<0.0f)?(OutAzimuth+2.0f*PI):OutAzimuth;
}


/*! This function constructs a linear combination of two Dirac-delta distributions 
   which matches three given moments.
  \param OutDepth The locations of the Dirac-delta distributions. They are sorted.
  \param OutWeight1 The weight of the second Dirac-delta. The weights add up to one.
  \param OptimizedMoments Four moments stored in a manner that is optimized for 
		 quantization. The fourth moment is not used.*/
void GetSparseRepresentation4Moments(out float2 OutDepth,out float OutWeight1,float4 OptimizedMoments){
	// Get the first three moments
	float3 b;
	b.xz=mul(OptimizedMoments.xz-0.5f,float2x2(-1.0f/3.0f,-0.75f,sqrt(3.0f),0.75f*sqrt(3.0f)));
	b.y=mul(OptimizedMoments.yw,float2x1(0.125f,1.0f));
	// Determine the locations of the Dirac-delta pulses
	float3 Polynomial;
	float Variance=mad(-b.x,b.x,b.y);
	Polynomial.z=Variance;
	Polynomial.y=mad(b.x,b.y,-b.z);
	Polynomial.x=dot(-b.xy,Polynomial.yz);
	float Scaling=1.0f/Polynomial[2];
	float p=Polynomial[1]*Scaling;
	float q=Polynomial[0]*Scaling;
	float D=p*p*0.25f-q;
	float r=sqrt(D);
	OutDepth=float2(-0.5f*p-r,-0.5f*p+r);
	// Get the weight
	OutWeight1=(b.x-OutDepth[0])/(OutDepth[1]-OutDepth[0]);
	// Negative variance may occur due to quantization errors. This is easily and 
	// appropriately handled using a single Dirac-delta at the mean
	bool SingleDiracDelta=(Variance<=1.0e-6f);
	OutDepth=SingleDiracDelta?float2(b.x,0.0f):OutDepth;
	OutWeight1=SingleDiracDelta?0.0f:OutWeight1;
}


/*! Estimates the integral of a probability distribution from minus infinity to a 
   fixed point and returns it.
  \param IntervalEnd The end of the integration interval.
  \param OptimizedMoments Power moments one to four of the distribution stored in a 
		 a manner that is optimized for little quantization errors.
  \param MomentBias The strength of the bias that serves to compensate quantization 
		 errors (e.g. 6*10^-5).
  \param OverestimationWeight If this value is zero, the function underestimates the 
		 integral. If it is one, it overestimates. Inbetween linear interpolation is 
		 employed.*/
float EstimateIntegralFrom4Moments(float IntervalEnd,float4 OptimizedMoments,float MomentBias,float OverestimationWeight){
	float4 _4Moments;
	_4Moments.xz=mul(OptimizedMoments.xz-0.5f,float2x2(-1.0f/3.0f,-0.75f,sqrt(3.0f),0.75f*sqrt(3.0f)));
	_4Moments.yw=mul(OptimizedMoments.yw,float2x2(0.125f,-0.125f,1.0f,1.0f));

	// Bias input data to avoid artifacts
	const float4 BiasVector=float4(0.0f,0.628f,0.0f,0.628f);
	float4 b=lerp(_4Moments,BiasVector,MomentBias);
	float3 z;
	z[0]=IntervalEnd;

	// Compute a Cholesky factorization of the Hankel matrix B storing only non-
	// trivial entries or related products
	float L21D11=mad(-b[0],b[1],b[2]);
	float D11=mad(-b[0],b[0], b[1]);
	float SquaredDepthVariance=mad(-b[1],b[1], b[3]);
	float D22D11=dot(float2(SquaredDepthVariance,-L21D11),float2(D11,L21D11));
	float InvD11=1.0f/D11;
	float L21=L21D11*InvD11;

	// Obtain a scaled inverse image of bz=(1,z[0],z[0]*z[0])^T
	float3 c=float3(1.0f,z[0],z[0]*z[0]);
	// Forward substitution to solve L*c1=bz
	c[1]-=b.x;
	c[2]-=b.y+L21*c[1];
	// Scaling to solve D*c2=c1
	c[1]*=InvD11;
	c[2]*=D11/D22D11;
	// Backward substitution to solve L^T*c3=c2
	c[1]-=L21*c[2];
	c[0]-=dot(c.yz,b.xy);
	// Solve the quadratic equation c[0]+c[1]*z+c[2]*z^2 to obtain solutions 
	// z[1] and z[2]
	float InvC2=1.0f/c[2];
	float p=c[1]*InvC2;
	float q=c[0]*InvC2;
	float D=(p*p*0.25f)-q;
	float r=sqrt(D);
	z[1]=-p*0.5f-r;
	z[2]=-p*0.5f+r;
	// Compute the shadow intensity by summing the appropriate weights
	float3 Weight;
	Weight[0]=(z[1]*z[2]-b[0]*(z[1]+z[2])+b[1])/((z[0]-z[1])*(z[0]-z[2]));
	Weight[1]=(z[0]*z[2]-b[0]*(z[0]+z[2])+b[1])/((z[2]-z[1])*(z[0]-z[1]));
	Weight[2]=1.0f-Weight[0]-Weight[1];
	float IntegralLowerBound=
		(z[2]<z[0])?(Weight[1]+Weight[2]):(
		(z[1]<z[0])?Weight[1]:0.0f);
	float IntegralUpperBound=saturate(IntegralLowerBound+Weight[0]);
	IntegralLowerBound=saturate(IntegralLowerBound);
	return lerp(IntegralLowerBound,IntegralUpperBound,OverestimationWeight);
}

/*! Given the coefficients of a real, cubic polynomial Coefficient[0]
  +x*Coefficient[1]+x*x*Coefficient[2]+x*x*x*Coefficient[3] this function computes 
  all of its roots and returns them.
 \warning Behavior is undefined if the polynomial does not have three real roots.
 \note The implementation is a cheap and specialized version of the algorithm 
	   proposed by James F. Blinn in his column "How to Solve a Cubic Equation". The 
	   full algorithm is summarized in part 5, published in IEEE Computer Graphics 
	   and Applications, vol. 27, no. 3.*/
float3 GetRoots(float4 Coefficient){
	// Since we do not want to have a homogenous representation of the roots in the 
	// end, we will need to divide by the leading coefficient. To save some other 
	// multiplications, we may just as well do this right now.
	Coefficient.xyz/=Coefficient.w;
	// Formulae are more handy if the coefficients in the middle are divided by 
	// three
	Coefficient.yz/=3.0f;
	// Compute the Hessian and the discrimant
	float3 Delta=float3(
		mad(-Coefficient.z,Coefficient.z,Coefficient.y),
		mad(-Coefficient.y,Coefficient.z,Coefficient.x),
		dot(float2(Coefficient.z,-Coefficient.y),Coefficient.xy)
	);
	float Discriminant=dot(float2(4.0f*Delta.x,-Delta.y),Delta.zy);
	// Compute coefficients of the depressed cubic (third is zero, fourth is one)
	float2 Depressed=float2(
		mad(-2.0f*Coefficient.z,Delta.x,Delta.y),
		Delta.x
	);
	// Take the cubic root of a normalized complex number
	float Theta=atan2(sqrt(Discriminant),-Depressed.x)/3.0f;
	float2 CubicRoot;
	sincos(Theta,CubicRoot.y,CubicRoot.x);
	// Compute the three roots, scale appropriately and revert the depression 
	// transform
	float3 Root=float3(
		CubicRoot.x,
		dot(float2(-0.5f,-0.5f*sqrt(3.0f)),CubicRoot),
		dot(float2(-0.5f, 0.5f*sqrt(3.0f)),CubicRoot)
	);
	Root=mad(2.0f*sqrt(-Depressed.y),Root,-Coefficient.z);
	return Root;
}

/*! Like EstimateIntegralFrom4Moments() but uses six moments which have been stored 
   in a manner that is optimized for small quantization error.*/
float EstimateIntegralFrom6Moments(float IntervalEnd,float2 OptimizedMoments0,float4 OptimizedMoments1,float MomentBias,float OverestimationWeight){
	// Dequantize the moments
	const float3x3 QuantizationMatrixOdd=float3x3(
			-0.02877789192f, 0.09995235706f, 0.25893353755f,
			 0.47635550422f, 0.84532580931f, 0.90779616657f,
			 1.55242808973f, 1.05472570761f, 0.83327335647f);
	const float3x3 QuantizationMatrixEven=float3x3(
			 0.00001253044f,-0.24998746956f,-0.37498825271f,
			 0.16668494186f, 0.16668494186f, 0.21876713299f,
			 0.86602540579f, 0.86602540579f, 0.81189881793f);
	float3 OptimizedMomentsOdd=float3(OptimizedMoments0,OptimizedMoments1.x)-0.5f;
	float3 OptimizedMomentsEven=OptimizedMoments1.yzw;
	OptimizedMomentsEven.z-=0.018888946f;
	float3 MomentsOdd=mul(OptimizedMomentsOdd,QuantizationMatrixOdd);
	float3 MomentsEven=mul(OptimizedMomentsEven,QuantizationMatrixEven);
	float b[6]={MomentsOdd.x,MomentsEven.x,MomentsOdd.y,MomentsEven.y,MomentsOdd.z,MomentsEven.z};
	// Bias input data to avoid artifacts
	const float pBiasVector[6]={0.0f,0.5566f,0.0f,0.489f,0.0f,0.47869382f};
	[unroll] for(int i=0;i!=6;++i){
		b[i]=lerp(b[i],pBiasVector[i],MomentBias);
	}
	float4 z;
	z[0]=IntervalEnd;

	// Compute a Cholesky factorization of the Hankel matrix B storing only non-
	// trivial entries or related products
	float InvD11=1.0f/mad(-b[0],b[0],b[1]);
	float L21D11=mad(-b[0],b[1],b[2]);
	float L21=L21D11*InvD11;
	float D22=mad(-L21D11,L21,mad(-b[1],b[1],b[3]));
	float L31D11=mad(-b[0],b[2],b[3]);
	float L31=L31D11*InvD11;
	float InvD22=1.0f/D22;
	float L32D22=mad(-L21D11,L31,mad(-b[1],b[2],b[4]));
	float L32=L32D22*InvD22;
	float D33=mad(-b[2],b[2],b[5])-dot(float2(L31D11,L32D22),float2(L31,L32));
	float InvD33=1.0f/D33;

	// Construct the polynomial whose roots have to be points of support of the 
	// canonical distribution: bz=(1,z[0],z[0]*z[0],z[0]*z[0]*z[0])^T
	float4 c;
	c[0]=1.0f;
	c[1]=z[0];
	c[2]=c[1]*z[0];
	c[3]=c[2]*z[0];
	// Forward substitution to solve L*c1=bz
	c[1]-=b[0];
	c[2]-=mad(L21,c[1],b[1]);
	c[3]-=b[2]+dot(float2(L31,L32),c.yz);
	// Scaling to solve D*c2=c1
	c.yzw*=float3(InvD11,InvD22,InvD33);
	// Backward substitution to solve L^T*c3=c2
	c[2]-=L32*c[3];
	c[1]-=dot(float2(L21,L31),c.zw);
	c[0]-=dot(float3(b[0],b[1],b[2]),c.yzw);

	// Solve the cubic equation c[0]+c[1]*z+c[2]*z^2+c[3]*z^3 to obtain solutions 
	// z[1], z[2] and z[3]
	z.yzw=GetRoots(c);

	// Determine how the weights at these roots contribute to the end result
	float4 WeightFactor;
	WeightFactor[0]=OverestimationWeight;
	WeightFactor.yzw=(z.yzw>z.xxx)?float3(0.0f,0.0f,0.0f):float3(1.0f,1.0f,1.0f);
	// Construct an interpolation polynomial matching these factors at the Roots
	float f0=WeightFactor[0];
	float f1=WeightFactor[1];
	float f2=WeightFactor[2];
	float f3=WeightFactor[3];
	float f01=(f1-f0)/(z[1]-z[0]);
	float f12=(f2-f1)/(z[2]-z[1]);
	float f23=(f3-f2)/(z[3]-z[2]);
	float f012=(f12-f01)/(z[2]-z[0]);
	float f123=(f23-f12)/(z[3]-z[1]);
	float f0123=(f123-f012)/(z[3]-z[0]);
	// Convert from the Newton basis to the canonical basis of polynomials
	float4 InterpolationPolynomial;
	// f012+f0123*(z-z2)
	InterpolationPolynomial[0]=mad(-f0123,z[2],f012);
	InterpolationPolynomial[1]=f0123;
	// *(z-z1)  +f01
	InterpolationPolynomial[2]=InterpolationPolynomial[1];
	InterpolationPolynomial[1]=mad(InterpolationPolynomial[1],-z[1],InterpolationPolynomial[0]);
	InterpolationPolynomial[0]=mad(InterpolationPolynomial[0],-z[1],f01);
	// *(z-z0)  +f0
	InterpolationPolynomial[3]=InterpolationPolynomial[2];
	InterpolationPolynomial[2]=mad(InterpolationPolynomial[2],-z[0],InterpolationPolynomial[1]);
	InterpolationPolynomial[1]=mad(InterpolationPolynomial[1],-z[0],InterpolationPolynomial[0]);
	InterpolationPolynomial[0]=mad(InterpolationPolynomial[0],-z[0],f0);

	// Now the end result happens to be the dot product of the first four moments 
	// and the interpolation polynomial
	return saturate(dot(InterpolationPolynomial,float4(1.0f,b[0],b[1],b[2])));
}


/*! This function computes the amount of light that is transmitted through one texel 
   of a rectified shadow map horizontally at depth zero.
  \param ShadowMapWidth The width of the shadow map in texels.
  \param ExtinctionCoefficient The extinction coefficient of the participating 
		 media.
  \param DistanceMinMax,InclinationMinMax See GetRectifiedSpaceFrustumBounds().*/
float ComputeTexelTransmittance(uint ShadowMapWidth,float ExtinctionCoefficient,float2 DistanceMinMax,float2 InclinationMinMax){
	float RaySegmentLength=DistanceMinMax.y/(sin(0.5f*(InclinationMinMax.x+InclinationMinMax.y))*float(ShadowMapWidth));
	return exp(-ExtinctionCoefficient*RaySegmentLength);
}


/*! This is a helper function for functions generating transmittance weighted shadow 
   maps with resampling of a shadow map for non-linear rectification. It precomputes 
   various useful quantities that can be prepared before entering a loop over a row 
   of a shadow map.
  \param OutTextureSpaceRayStart,OutTextureSpaceRayOffset Coordinates defining the 
		 ray which is to be traced in the source shadow map. OutTextureSpaceRayStart 
		 correspons to the left-most texel. They use projective coordinates, you 
		 should divide by z to get the texture coordinate in xy. 
  \param OutInclinationToDepthTransform A linear transform mapping inclination in 
		 radians to a value that can be used as depth value for the rectified shadow 
		 map.
  \param OutTexelTransmittance See ComputeTexelTransmittance().
  \param OutputShadowMapResolution The resolution of the shadow map in which prefix 
		 sums will be stored.
  \param iShadowMapRow The index of the row in the output shadow map (i.e. the 
		 non-negative integer y-coordinate).
  \param ExtinctionCoefficient The extinction coefficient of the participating 
		 media.
  \param RectificationToShadowMapProjectionSpace See 
		 ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4().
  \param DistanceMinMax,InclinationMinMax,AzimuthMinMax See 
		 GetRectifiedSpaceFrustumBounds().
  \sa ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4() */
void PrepareShadowMapRowResampling(out float3 OutTextureSpaceRayStart,out float3 OutTextureSpaceRayOffset,out float2 OutInclinationToDepthTransform,out float OutOffsetLength,out float OutInitialOffsetLength,out float OutTexelTransmittance,uint2 OutputShadowMapResolution,uint iShadowMapRow,float ExtinctionCoefficient,float4x4 RectificationToShadowMapProjectionSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	// Construct the reference ray in the epipolar slice
	float RayAzimuth=lerp(AzimuthMinMax.x,AzimuthMinMax.y,float(iShadowMapRow)/float(OutputShadowMapResolution.y));
	float ReferenceInclination=dot(InclinationMinMax,float2(0.5f,0.5f));
	float4 RectificationSpaceRayDirection=float4(RectifiedToCartesian(1.0f,ReferenceInclination,RayAzimuth),0.0f);
	float4 ProjectionSpaceRayDirection=mul(RectificationSpaceRayDirection,RectificationToShadowMapProjectionSpace);
	float4 ProjectionSpaceRayStart=mul(float4(0.0f,0.0f,0.0f,1.0f),RectificationToShadowMapProjectionSpace);
	// Convert to a coordinate system where shadow map texture coordinates arise 
	// directly to save one multiply-add within the loop
	float3 TextureSpaceRayDirection=ProjectionSpaceRayDirection.xyw*float3(0.5f,-0.5f,1.0f)+ProjectionSpaceRayDirection.www*float3(0.5f,0.5f,0.0f);
	// Also prepare coefficients for conversion of inclinations to depth
	OutInclinationToDepthTransform.x=1.0f/(InclinationMinMax.y-InclinationMinMax.x);
	OutInclinationToDepthTransform.y=-InclinationMinMax.x*OutInclinationToDepthTransform.x;
	OutInclinationToDepthTransform*=2.0f;
	OutInclinationToDepthTransform.y-=1.0f;
	// Compute the texel transmittance
	OutTexelTransmittance=ComputeTexelTransmittance(OutputShadowMapResolution.x,ExtinctionCoefficient,DistanceMinMax,InclinationMinMax);
	// Compute planar distances
	OutOffsetLength=DistanceMinMax.y/float(OutputShadowMapResolution.x);
	OutInitialOffsetLength=0.5f*OutOffsetLength;
	// Compute the offset length and the final offset
	OutTextureSpaceRayOffset=TextureSpaceRayDirection*OutOffsetLength;
	OutTextureSpaceRayStart=ProjectionSpaceRayStart.xyw*float3(0.5f,-0.5f,1.0f)+ProjectionSpaceRayStart.www*float3(0.5f,0.5f,0.0f);
	OutTextureSpaceRayStart+=0.5f*OutTextureSpaceRayOffset;
}


/*! This function converts a depth taken from a shadow map and maps it to a depth 
   value as used for non-linear rectification.
  \param ShadowMapDepth The depth value obtained from the shadow map mapped into the 
		 interval [-1,1].
  \param ShadowMapToRectificationDepthTransform,InclinationToDepthTransform See 
		 PrepareShadowMapRowResampling().
  \param TotalOffsetLength The number of times the offset has been applied to the 
		 ray start times the length of the offset as generated by 
		 PrepareShadowMapRowResampling().
  \return The resampled depth value. It is obtained from a shadow map sample at the 
		  given texture coordinate using a non-linear mapping.*/
float ConvertShadowMapDepthNonLinearRectification(float ShadowMapDepth,float2 ShadowMapToRectificationDepthTransform,float2 InclinationToDepthTransform,float TotalOffsetLength){
	float RectificationSpaceDepth=mad(ShadowMapDepth,ShadowMapToRectificationDepthTransform.x,ShadowMapToRectificationDepthTransform.y);
	float SampleInclination=atan2(TotalOffsetLength,RectificationSpaceDepth);
	return clamp(mad(SampleInclination,InclinationToDepthTransform.x,InclinationToDepthTransform.y),-1.0f,1.0f);
}


/*! Applies the optimized quantization transform to the given four power moments.*/
void Convert4MomentCanonicalToOptimized(out float4 OutQuantizedMoments0,float4 CanonicalMoments0){
	OutQuantizedMoments0.xz=mul(CanonicalMoments0.xz,float2x2(1.5f,sqrt(3.0f)*0.5f,-2.0f,-sqrt(3.0f)*2.0f/9.0f))+0.5f;
	OutQuantizedMoments0.yw=mul(CanonicalMoments0.yw,float2x2(4.0f,0.5f,-4.0f,0.5f));
}


/*! Converts a given depth value to a vector encoding four moments of a Dirac-delta 
   distribution at this depth. An affine transform is applied to the moments to 
   prepare them for storage in a texture with minimal loss of information due to 
   quantization.*/
void ComputeMomentVector4MomentsOptimized(out float4 OutQuantizedMoments0,float Depth){
	float SquaredDepth=Depth*Depth;
	Convert4MomentCanonicalToOptimized(OutQuantizedMoments0,float4(Depth,SquaredDepth,SquaredDepth*Depth,SquaredDepth*SquaredDepth));
}


/*! Applies the optimized quantization transform to the given six power moments.*/
void Convert6MomentCanonicalToOptimized(out float2 OutQuantizedMoments0,out float4 OutQuantizedMoments1,float2 CanonicalMoments0,float4 CanonicalMoments1){
	float3 MomentsOdd=float3(CanonicalMoments0.x,CanonicalMoments1.x,CanonicalMoments1.z);
	float3 MomentsEven=float3(CanonicalMoments0.y,CanonicalMoments1.y,CanonicalMoments1.w);
	// Apply the affine transform
	const float3x3 QuantizationMatrixOdd=float3x3(
			  2.5f,-1.87499864450f, 1.26583039016f,
			-10.0f, 4.20757543111f,-1.47644882902f,
			  8.0f,-1.83257678661f, 0.71061660238f);
	const float3x3 QuantizationMatrixEven=float3x3(
			 4.0f,  9.0f,-0.57759806484f,
			-4.0f,-24.0f, 4.61936647543f,
			 0.0f, 16.0f,-3.07953906655f);
	float3 OptimizedMomentsOdd=mul(MomentsOdd,QuantizationMatrixOdd)+0.5f;
	float3 OptimizedMomentsEven=mul(MomentsEven,QuantizationMatrixEven);
	OptimizedMomentsEven.z+=0.018888946f;
	OutQuantizedMoments0=OptimizedMomentsOdd.xy;
	OutQuantizedMoments1=float4(OptimizedMomentsOdd.z,OptimizedMomentsEven);
}


/*! Converts a given depth value to a vector encoding six moments of a Dirac-delta 
   distribution at this depth. The moments are output as two vectors and an affine 
   transform is applied to them to prepare them for storage in a texture with 
   minimal loss of information due to quantization.*/
void ComputeMomentVector6MomentsOptimized(out float2 OutQuantizedMoments0,out float4 OutQuantizedMoments1,float Depth){
	// Generate the moments
	float SquaredDepth=Depth*Depth;
	float CubedDepth=Depth*SquaredDepth;
	float2 CanonicalMoments0=float2(Depth,SquaredDepth);
	float4 CanonicalMoments1=float4(CubedDepth,SquaredDepth*SquaredDepth,CubedDepth*SquaredDepth,CubedDepth*CubedDepth);
	Convert6MomentCanonicalToOptimized(OutQuantizedMoments0,OutQuantizedMoments1,CanonicalMoments0,CanonicalMoments1);
}


#define TEXEL_TYPE float2
/*! This function generates a transmittance-weighted prefix sum for an already 
   rectified filterable shadow map. It only writes to every eights texel in its row 
   and thus implements a single thread of a thread group of width 8.
  \param OutFilteredShadowMap The texture to which the output is written.
  \param FilterableShadowMap The rectified, filterable shadow map from which the 
		 input is read. It must have resolution identical to OutFilteredShadowMap.
  \param iThreadX The thread will write to columns of the form 8*x+iThreadX where 
		 x iterates over all integers yielding a coordinate within the texture.
  \param iThreadY The index of the row from which the function will read and to 
		 which it will write.
  \param TexelTransmittance The source texel in column x is weighted by 
		 TexelTransmittance^x. Output values are divided by the total weight of the 
		 prefix sum. See ComputeTexelTransmittance().*/
void ApplyTransmittanceWeightedPrefixSum(inout RWTexture2D<TEXEL_TYPE> OutFilteredShadowMap,Texture2D<TEXEL_TYPE> FilterableShadowMap,uint iThreadX,uint iThreadY,float TexelTransmittance){
	const uint ThreadGroupWidth=8;
	uint Width,Height;
	OutFilteredShadowMap.GetDimensions(Width,Height);
	uint DividedWidth=Width/ThreadGroupWidth;
	// Prepare running sums and products for the loop
	float Weight=1.0f;
	float TotalWeight=0.0f;
	TEXEL_TYPE PrefixSum;
	PrefixSum=0.0f;
	// Iterate over the texture block by block
	[loop] for(uint x=0;x<DividedWidth;++x){
		uint BlockX=x*ThreadGroupWidth;
		TEXEL_TYPE StoredPrefixSum=PrefixSum;
		float StoredTotalWeight=TotalWeight;
		[unroll] for(uint i=0;i!=ThreadGroupWidth;++i){
			// We let each thread load all texture samples for its row. This means 
			// that we have ThreadGroupWidth as many texture reads as we would need 
			// to have but parallelism is increased by the same factor and 
			// benchmarks show that it is still very fast thanks to texture cache. 
			// In fact, the observed speed is better than caching via group-shared 
			// memory.
			PrefixSum+=Weight*FilterableShadowMap.Load(uint3(BlockX+i,iThreadY,0));
			TotalWeight+=Weight;
			Weight*=TexelTransmittance;
			bool StoreResult=(i==iThreadX);
			StoredPrefixSum=StoreResult?PrefixSum:StoredPrefixSum;
			StoredTotalWeight=StoreResult?TotalWeight:StoredTotalWeight;
		}
		OutFilteredShadowMap[uint2(BlockX+iThreadX,iThreadY)]=StoredPrefixSum/StoredTotalWeight;
	}
}


#define TEXEL_TYPE float3
void ApplyTransmittanceWeightedPrefixSum(inout RWTexture2D<TEXEL_TYPE> OutFilteredShadowMap,Texture2D<TEXEL_TYPE> FilterableShadowMap,uint iThreadX,uint iThreadY,float TexelTransmittance){
	const uint ThreadGroupWidth=8;
	uint Width,Height;
	OutFilteredShadowMap.GetDimensions(Width,Height);
	uint DividedWidth=Width/ThreadGroupWidth;
	// Prepare running sums and products for the loop
	float Weight=1.0f;
	float TotalWeight=0.0f;
	TEXEL_TYPE PrefixSum;
	PrefixSum=0.0f;
	// Different threads handle different textures
	[loop] for(uint x=0;x<DividedWidth;++x){
		uint BlockX=x*ThreadGroupWidth;
		TEXEL_TYPE StoredPrefixSum=PrefixSum;
		float StoredTotalWeight=TotalWeight;
		[unroll] for(uint i=0;i!=ThreadGroupWidth;++i){
			// We let each thread load all texture samples for its row. This means 
			// that we have ThreadGroupWidth as many texture reads as we would need 
			// to have but parallelism is increased by the same factor and 
			// benchmarks show that it is still very fast thanks to texture cache. 
			// In fact, the observed speed is better than caching via group-shared 
			// memory.
			PrefixSum+=Weight*FilterableShadowMap.Load(uint3(BlockX+i,iThreadY,0));
			TotalWeight+=Weight;
			Weight*=TexelTransmittance;
			bool StoreResult=(i==iThreadX);
			StoredPrefixSum=StoreResult?PrefixSum:StoredPrefixSum;
			StoredTotalWeight=StoreResult?TotalWeight:StoredTotalWeight;
		}
		OutFilteredShadowMap[uint2(BlockX+iThreadX,iThreadY)]=StoredPrefixSum/StoredTotalWeight;
	}
}


#define TEXEL_TYPE float4
void ApplyTransmittanceWeightedPrefixSum(inout RWTexture2D<TEXEL_TYPE> OutFilteredShadowMap,Texture2D<TEXEL_TYPE> FilterableShadowMap,uint iThreadX,uint iThreadY,float TexelTransmittance){
	const uint ThreadGroupWidth=8;
	uint Width,Height;
	OutFilteredShadowMap.GetDimensions(Width,Height);
	uint DividedWidth=Width/ThreadGroupWidth;
	// Prepare running sums and products for the loop
	float Weight=1.0f;
	float TotalWeight=0.0f;
	TEXEL_TYPE PrefixSum;
	PrefixSum=0.0f;
	// Different threads handle different textures
	[loop] for(uint x=0;x<DividedWidth;++x){
		uint BlockX=x*ThreadGroupWidth;
		TEXEL_TYPE StoredPrefixSum=PrefixSum;
		float StoredTotalWeight=TotalWeight;
		[unroll] for(uint i=0;i!=ThreadGroupWidth;++i){
			// We let each thread load all texture samples for its row. This means 
			// that we have ThreadGroupWidth as many texture reads as we would need 
			// to have but parallelism is increased by the same factor and 
			// benchmarks show that it is still very fast thanks to texture cache. 
			// In fact, the observed speed is better than caching via group-shared 
			// memory.
			PrefixSum+=Weight*FilterableShadowMap.Load(uint3(BlockX+i,iThreadY,0));
			TotalWeight+=Weight;
			Weight*=TexelTransmittance;
			bool StoreResult=(i==iThreadX);
			StoredPrefixSum=StoreResult?PrefixSum:StoredPrefixSum;
			StoredTotalWeight=StoreResult?TotalWeight:StoredTotalWeight;
		}
		OutFilteredShadowMap[uint2(BlockX+iThreadX,iThreadY)]=StoredPrefixSum/StoredTotalWeight;
	}
}
