/*! \file
   This header defines utility functions related to the trigonometric moment 
   problem and its applications for shadow mapping.*/

#include "TrigonometricMomentProblemShadowAlgebra.fx"

/*! Trigonometric moments require pi.*/
static const float PI=3.1415926535897932384626433832795f;

/*! @{
	Converts one or more points on the unit-circle to parameters identifying the 
	location of these points on the circle. The parameter is guaranteed to increase 
	monotonously as the point goes away from (1.0,0.0) in counter-clockwise 
	direction.
  \warning Except for the mentioned guarantee no guarantees are made. In particular 
		   this function does not return an arclength parametrization! It should 
		   only be used to induce an order on points on the unit circle.*/
float CircleToParameter(float2 UnitCirclePoint){
	return (UnitCirclePoint.y>=0.0f)?(-UnitCirclePoint.x):(2.0f+UnitCirclePoint.x);
}
float2 CircleToParameter(float2x2 UnitCirclePoint){
	return (UnitCirclePoint._m01_m11>=0.0f)?(-UnitCirclePoint._m00_m10):(2.0f+UnitCirclePoint._m00_m10);
}
float3 CircleToParameter(float3x2 UnitCirclePoint){
	return (UnitCirclePoint._m01_m11_m21>=0.0f)?(-UnitCirclePoint._m00_m10_m20):(2.0f+UnitCirclePoint._m00_m10_m20);
}
float4 CircleToParameter(float4x2 UnitCirclePoint){
	return (UnitCirclePoint._m01_m11_m21_m31>=0.0f)?(-UnitCirclePoint._m00_m10_m20_m30):(2.0f+UnitCirclePoint._m00_m10_m20_m30);
}
void CircleToParameter(out float pOutParameter[5],float2 pUnitCirclePoint[5]){
	[unroll] for(int i=0;i!=5;++i){
		pOutParameter[i]=((pUnitCirclePoint[i].y>=0.0f)?(-pUnitCirclePoint[i].x):(2.0f+pUnitCirclePoint[i].x));
	}
}
//!@}


/*! This function computes the weight of the given distribution on the unit circle 
	segment from one to root zero of the given distribution in counter-clockwise 
	direction. Endpoints are excluded. Distributions are supposed to be given in the 
	form in which Get2TMSMCanonicalDistribution() returns them.*/
float GetWeightExclusive(float3 pWeight,float3x2 pRoot){
	float3 pParameter=CircleToParameter(pRoot);
	return dot((pParameter.yz<pParameter[0])?pWeight.yz:float2(0.0f,0.0f),float2(1.0f,1.0f));
}


/*! The function generating two trigonometric moments and the zeroth moment, i.e. 
   row j is exp(2*pi*i*j*Point).*/
float3x2 MomentGeneratingFunction2(float Point){
	float3x2 MomentVector;
	MomentVector[0]=float2(1.0f,0.0f);
	sincos(2.0f*PI*Point,MomentVector[1].y,MomentVector[1].x);
	MomentVector[2]=Square(MomentVector[1]);
	return MomentVector;
}


/*! Like MomentGeneratingFunction2(float) but expects the point on the unit circle 
   instead of its scaled argument. Row j is CirclePoint^j.*/
float3x2 MomentGeneratingFunction2(float2 CirclePoint){
	float3x2 MomentVector;
	MomentVector[0]=float2(1.0f,0.0f);
	MomentVector[1]=CirclePoint;
	MomentVector[2]=Square(MomentVector[1]);
	return MomentVector;
}


/*! This function constructs the unique canonical distribution matching the given 
   trigonometric moments and having the given root. It is assumed that the first and 
   second moment are given and that the zeroth moment is one. It outputs the weights 
   and the corresponding roots. The roots are given as points on the unit circle. 
   The prescribed root has index zero.*/
void Get2TMSMCanonicalDistribution(out float3 pOutWeight,out float3x2 pOutRoot,float2 pTrigonometricMoment[2],float PrescribedRoot){
	// Gather inputs in convenient ways
	float2 b1=pTrigonometricMoment[0];
	float2 b2=pTrigonometricMoment[1];
	// Compute the right-hand side of the first linear system
	float3x2 bz0=MomentGeneratingFunction2(PrescribedRoot);
	// Construct the quadratic polynomial that has points of support as roots
	float3x2 c=Conjugate(SolveToeplitzSystem(b1,b2,bz0[1],bz0[2]));
	// Solve the polynomial equation to get all roots of the distribution
	float2x2 pRoot=GetRoots(c);
	pOutRoot[0]=bz0[1];
	pOutRoot[1]=pRoot[0];
	pOutRoot[2]=pRoot[1];
	// Compute the weights
	float InvDenominator=1.0f/dot(pOutRoot._m21_m01_m11-pOutRoot._m11_m21_m01,pOutRoot._m00_m10_m20);
	pOutWeight[0]= dot(float3(pOutRoot[1].y-b1.y,b1.y-pOutRoot[2].y,pOutRoot[2].y-pOutRoot[1].y),float3(pOutRoot[2].x,pOutRoot[1].x,b1.x))*InvDenominator;
	pOutWeight[1]=-dot(float3(pOutRoot[0].y-b1.y,b1.y-pOutRoot[2].y,pOutRoot[2].y-pOutRoot[0].y),float3(pOutRoot[2].x,pOutRoot[0].x,b1.x))*InvDenominator;
	pOutWeight[2]=1.0f-pOutWeight[0]-pOutWeight[1];
}


/*! Computes the canonical distribution for the given moments and root (see 
   Get2TMSMCanonicalDistribution()) and returns its total weight on the interval 
   from 0.0 to PrescribedRoot excluding endpoints.*/
float Get2TMSMCanonicalWeightExclusive(float2 pTrigonometricMoment[2],float PrescribedRoot){
	// Compute roots of the canonical representation
	float3x2 bz0=MomentGeneratingFunction2(PrescribedRoot);
	float2x2 pAdditionalRoot=GetRoots(Conjugate(SolveToeplitzSystem(pTrigonometricMoment[0],pTrigonometricMoment[1],bz0[1],bz0[2])));
	// Sort the roots
	float2 pAdditionalParameter=CircleToParameter(pAdditionalRoot);
	bool SwapRoots=(pAdditionalParameter.x>pAdditionalParameter.y);
	float3x2 pRoot;
	pRoot[0]=bz0[1];
	pRoot[1]=(SwapRoots?pAdditionalRoot[1]:pAdditionalRoot[0]);
	pRoot[2]=(SwapRoots?pAdditionalRoot[0]:pAdditionalRoot[1]);
	pAdditionalParameter=SwapRoots?pAdditionalParameter.yx:pAdditionalParameter.xy;
	// Figure out which combination of weights is needed
	float PrescribedParameter=CircleToParameter(pRoot[0]);
	float4 Switch=
		(pAdditionalParameter.y<PrescribedParameter)?float4(pRoot[1],1.0f,1.0f):(
		(pAdditionalParameter.x<PrescribedParameter)?float4(pRoot[0],0.0f,1.0f):
		float4(0.0f,0.0f,0.0f,0.0f));
	// Evaluate the combination of weights
	float2 b1=pTrigonometricMoment[0];
	float Quotient=dot(float3(Switch.y-b1.y,b1.y-pRoot[2].y,pRoot[2].y-Switch.y),float3(pRoot[2].x,Switch.x,b1.x))
				  /dot(pRoot._m21_m01_m11-pRoot._m11_m21_m01,pRoot._m00_m10_m20);
	return Switch[2]-Switch[3]*Quotient;
}


/*! Computes various sizes which are needed by multiple functions dealing with 
   optimization problems constrained by two normalized trigonometric moments.*/
void GetExtremalWeightHelpers2(out float3x2 OutMomentEnd,out float3x2 OutInvertedMomentEnd,out float3x2 OutInvertedMomentOne,out float2 OutEndDotOne,out float OutOneDotOne,float2 pTrigonometricMoment[2],float IntervalEnd){
	// Evaluate the moment generating function at IntervalEnd and compute its 
	// inverse with respect to the Toeplitz matrix
	OutMomentEnd=MomentGeneratingFunction2(IntervalEnd);
	OutInvertedMomentEnd=SolveToeplitzSystem(pTrigonometricMoment[0],pTrigonometricMoment[1],OutMomentEnd[1],OutMomentEnd[2]);
	// Do the same for the moment generating function at one
	OutInvertedMomentOne=SolveToeplitzSystem(pTrigonometricMoment[0],pTrigonometricMoment[1],float2(1.0f,0.0f),float2(1.0f,0.0f));
	// Compute various dot products, which are needed for both remaining cases
	OutEndDotOne=Conjugate(OutInvertedMomentEnd[0]+OutInvertedMomentEnd[1]+OutInvertedMomentEnd[2]);
	OutOneDotOne=dot(OutInvertedMomentOne._m00_m10_m20,float3(1.0f,1.0f,1.0f));
}


/*! This function computes critical points of a special weight map. If one of these 
   points is used as point of support for a distribution matching the given 
   normalized, trigonometric moments with weight at one, IntervalEnd and at most one 
   additional point, the weight at the returned point is either undefined or 
   extremal. The critical points are returned as points on the unit circle. The 
   output of GetExtremalWeightHelpers2() has to be provided.*/
float4x2 GetWeightMapCriticalPoints(float3x2 MomentEnd,float3x2 InvertedMomentEnd,float2 EndDotOne,float OneDotOne,float2 pTrigonometricMoment[2],float IntervalEnd){
	// Construct the polynomial
	float2 pCoefficient[5]={float2(0.0f,0.0f),float2(0.0f,0.0f),float2(0.0f,0.0f),float2(0.0f,0.0f),float2(0.0f,0.0f)};
	// Compute the inverse Toeplitz matrix
	float3x3 ToeplitzInverseReal,ToeplitzInverseImaginary;
	GetToeplitzInverse(ToeplitzInverseReal,ToeplitzInverseImaginary,pTrigonometricMoment);
	// Start with the terms, which do not include the interval ends, but need 
	// to be multiplied by EndDotOne afterwards
	[unroll] for(int i=0;i!=3;++i){
		[unroll] for(int j=0;j!=3;++j){
			float2 InverseToeplitzEntry=float2(ToeplitzInverseReal[i][j],ToeplitzInverseImaginary[i][j]);
			pCoefficient[2+j-i]+=(j-i)*InverseToeplitzEntry;
		}
	}
	// Multiply the first part by EndDotOne
	for(int i=0;i!=5;++i){
		pCoefficient[i]=Multiply(pCoefficient[i],EndDotOne);
	}
	// Invert the moment vector for the other interval end
	float3x2 InvertedMomentOne;
	InvertedMomentOne._m00_m10_m20=mul(ToeplitzInverseReal,float3(1.0f,1.0f,1.0f));
	InvertedMomentOne._m01_m11_m21=mul(ToeplitzInverseImaginary,float3(1.0f,1.0f,1.0f));
	// Construct the parts of the polynomial, which are formed by outer products 
	// of vectors depending on the variable, multiplied by inverted interval end 
	// vectors
	[unroll] for(int i=0;i!=3;++i){
		[unroll] for(int j=0;j!=3;++j){
			float2 EndTimesOne=Multiply(InvertedMomentOne[i],Conjugate(InvertedMomentEnd[j]));
			pCoefficient[2+j-i]-=(j-i)*EndTimesOne;
		}
	}
	// Compute the roots of the polynomial to obtain critical points of the 
	// weight map
	float4x2 pCriticalPoint=GetRoots(pCoefficient);
	// The critical points are known to lie on the unit circle, but due to numerical 
	// inaccurracies this may not hold in practice. Renormalize.
	[unroll] for(int i=0;i!=4;++i){
		pCriticalPoint[i]=normalize(pCriticalPoint[i]);
	}
	return pCriticalPoint;
}


/*! This function determines the weight that a distribution realizing the given 
   normalized, trigonometric moments should have at one if it is to minimize the 
   cumulative probability between zero and the given location.*/
float GetMinimizingWeightAtOne(float2 pTrigonometricMoment[2],float IntervalEnd){
	// Compute the canonical distribution at one
	float3x2 pRootOne;
	float3 pWeightOne;
	Get2TMSMCanonicalDistribution(pWeightOne,pRootOne,pTrigonometricMoment,1.0f);
	float3 pRootOneParameter=CircleToParameter(pRootOne);
	// Sort the two free roots of this canonical distribution
	bool SwapRootsOne=(pRootOneParameter[2]<pRootOneParameter[1]);
	pRootOneParameter.yz=SwapRootsOne?pRootOneParameter.zy:pRootOneParameter.yz;
	pWeightOne.yz=SwapRootsOne?pWeightOne.zy:pWeightOne.yz;
	// Determine the present interval
	float2 SegmentEnd;
	sincos(2.0f*PI*IntervalEnd,SegmentEnd.y,SegmentEnd.x);
	float IntervalEndParameter=CircleToParameter(SegmentEnd);
	// Any canonical distribution will not have a root between 0.0 and its 
	// prescribed root IntervalEnd, so any weight at one is fine. We use 0.0 for 
	// reasons of numerical stability.
	if(IntervalEndParameter<pRootOneParameter[1]){
		return 0.0f;
	}

	// Compute some shared, intermediate values
	float3x2 MomentEnd,InvertedMomentEnd,InvertedMomentOne;
	float2 EndDotOne;
	float OneDotOne;
	GetExtremalWeightHelpers2(MomentEnd,InvertedMomentEnd,InvertedMomentOne,EndDotOne,OneDotOne,pTrigonometricMoment,IntervalEnd);

	// We need to maximize the weight at the only root in the relevant interval. 
	// This can be done by solving a quartic equation for potentially optimal 
	// locations of this root.
	if(IntervalEndParameter<pRootOneParameter[2]){
		// Get critical points of the relevant weight map
		float4x2 pCriticalPoint=GetWeightMapCriticalPoints(MomentEnd,InvertedMomentEnd,EndDotOne,OneDotOne,pTrigonometricMoment,IntervalEnd);
		// Compute the canonical distribution at IntervalEnd to obtain the interval 
		// containing the relevant critical point
		float3x2 pRootEnd;
		float3 pWeightEnd;
		Get2TMSMCanonicalDistribution(pWeightEnd,pRootEnd,pTrigonometricMoment,IntervalEnd);
		float3 pRootEndParameter=CircleToParameter(pRootEnd);
		bool SwapRootsEnd=(pRootEndParameter[2]<pRootEndParameter[1]);
		float MinFirstRootParameter=SwapRootsEnd?pRootEndParameter[2]:pRootEndParameter[1];
		float MinFirstWeight=SwapRootsEnd?pWeightEnd[2]:pWeightEnd[1];
		// Find the critical point in this interval
		float2 FirstRoot;
		bool RootFound=false;
		float4 pCriticalPointParameter=CircleToParameter(pCriticalPoint);
		[unroll] for(int i=0;i!=4;++i){
			if(MinFirstRootParameter<=pCriticalPointParameter[i] && pCriticalPointParameter[i]<=pRootOneParameter[1]){
				FirstRoot=pCriticalPoint[i];
				RootFound=true;
			}
		}
		if(!RootFound){
			// There is no extremum in the relevant interval, so the correct 
			// solution has to be to use one of the two already computed canonical 
			// distributions. Find out which one.
			return (MinFirstWeight>pWeightOne[1])?pWeightOne[0]:0.0f;
		}
		// Compute the corresponding weight at one
		float3x2 MomentFirst;
		MomentFirst[0]=float2(1.0f,0.0f);
		MomentFirst[1]=FirstRoot;
		MomentFirst[2]=Square(FirstRoot);
		float2 FirstDotEnd=Dot(MomentFirst,InvertedMomentEnd);
		float2 FirstDotOne=Dot(MomentFirst,InvertedMomentOne);
		float Weight=Divide(FirstDotEnd,OneDotOne*FirstDotEnd-Multiply(FirstDotOne,Conjugate(EndDotOne))).x;
		// Clamp and return it
		return max(0.0f,min(pWeightOne[0],Weight));
	}
	// We need to maximize the sum of the weight at one and at IntervalEnd. This can 
	// be done by means of a rather explicit closed-form.
	else{
		float EndDotEnd=Dot(MomentEnd,InvertedMomentEnd).x;
		float Weight=(Magnitude(EndDotOne)-EndDotEnd)/(dot(EndDotOne,EndDotOne)-OneDotOne*EndDotEnd);
		// Clamp and return it
		return max(0.0f,min(pWeightOne[0],Weight));
	}
}


/*! This function computes a distribution realizing the given normalized, 
   trigonometric moments and having a minimal weight on the open interval 
   (0,IntervalEnd). This minimal weight is returned.*/
float Get2TMSMMinimalWeight(float2 pTrigonometricMoment[2],float IntervalEnd){
	// Compute the weight at one
	float WeightAtOne=GetMinimizingWeightAtOne(pTrigonometricMoment,IntervalEnd);
	// Update the moment vector and renormalize it
	float2 pModifiedMoment[2]={
		(pTrigonometricMoment[0]-float2(WeightAtOne,0.0f))/(1.0f-WeightAtOne),
		(pTrigonometricMoment[1]-float2(WeightAtOne,0.0f))/(1.0f-WeightAtOne)};
	// Compute the weight of the corresponding canonical distribution
	float CanonicalWeight=Get2TMSMCanonicalWeightExclusive(pModifiedMoment,IntervalEnd);
	// Scale it down to account for the weight at one
	return saturate(CanonicalWeight*(1.0f-WeightAtOne));
}
