/*! \file
   This module provides functions that are used by CConstantManager to define 
   redundant constants.*/

/*! This function computes the multiplicative inverse of a matrix whose last column 
   is 0 0 0 1.*/
void InvertMatrix(out float4x4 OutBToA,float4x4 AToB){
	float InvDeterminant=AToB._11*(AToB._22*AToB._33-AToB._32*AToB._23)
						-AToB._12*(AToB._21*AToB._33-AToB._31*AToB._23)
						+AToB._13*(AToB._21*AToB._32-AToB._31*AToB._22);
	// Invert the 3*3 matrix
	InvDeterminant=1.0f/InvDeterminant;
	OutBToA._11=(AToB._22*AToB._33-AToB._23*AToB._32)*InvDeterminant;
	OutBToA._12=(AToB._12*AToB._33-AToB._32*AToB._13)*-InvDeterminant;
	OutBToA._13=(AToB._12*AToB._23-AToB._22*AToB._13)*InvDeterminant;
	OutBToA._21=(AToB._21*AToB._33-AToB._23*AToB._31)*-InvDeterminant;
	OutBToA._22=(AToB._11*AToB._33-AToB._31*AToB._13)*InvDeterminant;
	OutBToA._23=(AToB._11*AToB._23-AToB._21*AToB._13)*-InvDeterminant;
	OutBToA._31=(AToB._21*AToB._32-AToB._31*AToB._22)*InvDeterminant;
	OutBToA._32=(AToB._11*AToB._32-AToB._31*AToB._12)*-InvDeterminant;
	OutBToA._33=(AToB._11*AToB._22-AToB._12*AToB._21)*InvDeterminant;
	// Invert the translation of the matrix
	OutBToA._41=-(AToB._41*OutBToA._11+AToB._42*OutBToA._21+AToB._43*OutBToA._31);
	OutBToA._42=-(AToB._41*OutBToA._12+AToB._42*OutBToA._22+AToB._43*OutBToA._32);
	OutBToA._43=-(AToB._41*OutBToA._13+AToB._42*OutBToA._23+AToB._43*OutBToA._33);
	// Ensure that all entries are set
	OutBToA._14=OutBToA._24=OutBToA._34=0.0f;
	OutBToA._44=1.0f;
}


/*! This function transforms one matrix by another matrix.*/
void TransformMatrix(out float4x4 OutAToCSpace,float4x4 AToBSpace,float4x4 BToCSpace){
	OutAToCSpace=mul(AToBSpace,BToCSpace);
}


/*! This function computes a uniform scaling value that matches the scaling in the 
   given matrix best and returns it.*/
void ComputeAverageScaling(out float OutAverageScaling,float4x4 Matrix){
	float3 Scaling=float3(dot(Matrix[0],Matrix[0]),dot(Matrix[1],Matrix[1]),dot(Matrix[2],Matrix[2]));
	Scaling=sqrt(Scaling);
	OutAverageScaling=(Scaling.x+Scaling.y+Scaling.z)/3.0f;
}


/*! Given a world to projection space transformation this function computes the six 
   clipping planes and normalizes them.*/
void ComputeClippingPlanes(out float4 OutLeft,out float4 OutRight,out float4 OutTop,out float4 OutBottom,out float4 OutNear,out float4 OutFar,float4x4 WorldToProjectionSpace){
	// The left clipping plane
	OutLeft=normalize(float4(
		WorldToProjectionSpace._14+WorldToProjectionSpace._11,
		WorldToProjectionSpace._24+WorldToProjectionSpace._21,
		WorldToProjectionSpace._34+WorldToProjectionSpace._31,
		WorldToProjectionSpace._44+WorldToProjectionSpace._41));
	// The right clipping plane
	OutRight=normalize(float4(
		WorldToProjectionSpace._14-WorldToProjectionSpace._11,
		WorldToProjectionSpace._24-WorldToProjectionSpace._21,
		WorldToProjectionSpace._34-WorldToProjectionSpace._31,
		WorldToProjectionSpace._44-WorldToProjectionSpace._41));
	// The top clipping plane
	OutTop=normalize(float4(
		WorldToProjectionSpace._14-WorldToProjectionSpace._12,
		WorldToProjectionSpace._24-WorldToProjectionSpace._22,
		WorldToProjectionSpace._34-WorldToProjectionSpace._32,
		WorldToProjectionSpace._44-WorldToProjectionSpace._42));
	// The bottom clipping plane
	OutBottom=normalize(float4(
		WorldToProjectionSpace._14+WorldToProjectionSpace._12,
		WorldToProjectionSpace._24+WorldToProjectionSpace._22,
		WorldToProjectionSpace._34+WorldToProjectionSpace._32,
		WorldToProjectionSpace._44+WorldToProjectionSpace._42));
	// The near clipping plane
	OutNear=normalize(float4(
		WorldToProjectionSpace._13,
		WorldToProjectionSpace._23,
		WorldToProjectionSpace._33,
		WorldToProjectionSpace._43));
	// The far clipping plane
	OutFar=normalize(float4(
		WorldToProjectionSpace._14-WorldToProjectionSpace._13,
		WorldToProjectionSpace._24-WorldToProjectionSpace._23,
		WorldToProjectionSpace._34-WorldToProjectionSpace._33,
		WorldToProjectionSpace._44-WorldToProjectionSpace._43));
}


/*! This function computes a normalized vector, which is orthogonal to the near 
   clipping plane and points out of the view frustum (a normal vector for 
   billboards). Besides it computes a tangent and a bitangent for billboards and 
   scaled versions of these values. */
void ComputeBillboardData(out float3 OutBillboardNormal,out float3 BillboardUpVector,out float3 BillboardRightVector,out float3 BillboardScaledUpVector,out float3 BillboardScaledRightVector,float2 BillboardSize,float4x4 WorldToViewSpace){
	// Since the world to view space transformation always has an orthogonal 3x3 
	// part its transpose can be used instead of the view to world space 
	// transformation
	OutBillboardNormal=float3(-WorldToViewSpace._13,-WorldToViewSpace._23,-WorldToViewSpace._33);
	OutBillboardRightVector=float3(WorldToViewSpace._11,WorldToViewSpace._21,WorldToViewSpace._31);
	OutBillboardUpVector=float3(WorldToViewSpace._12,WorldToViewSpace._22,WorldToViewSpace._32);
	OutBillboardScaledRightVector=OutBillboardRightVector*BillboardSize.x;
	OutBillboardScaledUpVector=OutBillboardUpVector*BillboardSize.y;
}


/*! This function computes the world space position of the camera.*/
void ComputeWorldSpaceCameraPosition(out float3 OutWorldSpaceCameraPosition,float4x4 ViewToWorldSpace){
	OutWorldSpaceCameraPosition=float3(ViewToWorldSpace._41,ViewToWorldSpace._42,ViewToWorldSpace._43);
}

/*! This function computes the mesh space position of the camera.*/
void ComputeMeshSpaceCameraPosition(out float3 OutMeshSpaceCameraPosition,float4x4 ViewToMeshSpace){
	OutMeshSpaceCameraPosition=float3(ViewToMeshSpace._41,ViewToMeshSpace._42,ViewToMeshSpace._43);
}
