#include "ShadowUtility.fx"
#include "TrigonometricMomentProblemShadow.fx"
#include "RandomNumbers.fx"

/*! \file
   Provides function definitions for all available shadow mapping techniques. This 
   includes parametrization, generation, post-processing and usage of shadow maps.*/


/*! Given the projection space position of a fragment this function outputs a 
   perspective depth value, which can be used to represent this fragment in a shadow 
   map.*/
void ComputeFragmentDepth(out float OutFragmentDepth,float4 ProjectionSpacePosition){
	OutFragmentDepth=saturate(ProjectionSpacePosition.z/ProjectionSpacePosition.w)*2.0f-1.0f;
}

/*! Maps the given fragment depth from the interval [-1,1] to [0,1] such that it can 
   be stored in an unsigned shadow map. This is to ensure consistency with shadow 
   maps created as depth buffers.*/
void ComputeShadowMapDepth(out float OutShadowMapDepth,float FragmentDepth){
	OutShadowMapDepth=FragmentDepth*0.5f+0.5f;
}

/*! Returns a vector where r contains Vector, gb contains zero and a contains 
   Alpha.*/
void CombineVectorWithAlpha_float1(out float4 OutVectorWithAlpha,float Vector,float Alpha){
	OutVectorWithAlpha=float4(Vector,0.0f,0.0f,Alpha);
}
/*! Returns a vector where rg contains Vector, b contains zero and a contains 
   Alpha.*/
void CombineVectorWithAlpha_float2(out float4 OutVectorWithAlpha,float2 Vector,float Alpha){
	OutVectorWithAlpha=float4(Vector,0.0f,Alpha);
}
/*! Returns a vector where rgb contains Vector and a contains Alpha.*/
void CombineVectorWithAlpha_float3(out float4 OutVectorWithAlpha,float3 Vector,float Alpha){
	OutVectorWithAlpha=float4(Vector,Alpha);
}
/*! Outputs Vector unmodified. There is no space for an alpha channel.*/
void CombineVectorWithAlpha_float4(out float4 OutVectorWithAlpha,float4 Vector){
	OutVectorWithAlpha=Vector;
}


/*! This function turns a shadow intensity into a color that can be displayed for 
   debugging purposes.*/
void ConvertShadowIntensityToColor(out float4 OutColor,float ShadowIntensity){
	OutColor.rgb=1.0f-ShadowIntensity;
	OutColor.a=1.0f;
}


/*! Given a scalar this function outputs a vector whose first component is the 
   exponential of the given depth times -Factor.*/
void ComputeMomentVectorExponential(out float OutExponential,float FragmentDepth,float Factor=80.0f){
	OutExponential=exp(Factor*FragmentDepth);
}

/*! Given a scalar this function outputs four vectors, each consisting of two 
   complex Fourier coefficients of the delta distribution realizing this scalar.*/
void ComputeMomentVectorFourier8(out float4 OutFourier1_2,out float4 OutFourier3_4,out float4 OutFourier5_6,out float4 OutFourier7_8,float FragmentDepth){
	const float PI=3.1415926535897932384626433832795f;
	// Compute the first Fourier coefficient
	float2 Fourier1;
	sincos((FragmentDepth*0.5f+0.5f)*PI,Fourier1.y,Fourier1.x);
	// Compute the other coefficients by complex exponentiation
	float2 FourierSquare=Multiply(Fourier1,Fourier1);
	float2 pFourierCoefficient[8];
	pFourierCoefficient[0]=Fourier1;
	[unroll] for(int i=1;i!=8;++i){
		pFourierCoefficient[i]=Multiply(pFourierCoefficient[i-1],FourierSquare);
	}
	// Store the results
	const float2 Factor=float2(0.5f,0.5f);
	const float2 Summand=Factor;
	OutFourier1_2.xy=pFourierCoefficient[0]*Factor+Summand;
	OutFourier1_2.zw=pFourierCoefficient[1]*Factor+Summand;
	OutFourier3_4.xy=pFourierCoefficient[2]*Factor+Summand;
	OutFourier3_4.zw=pFourierCoefficient[3]*Factor+Summand;
	OutFourier5_6.xy=pFourierCoefficient[4]*Factor+Summand;
	OutFourier5_6.zw=pFourierCoefficient[5]*Factor+Summand;
	OutFourier7_8.xy=pFourierCoefficient[6]*Factor+Summand;
	OutFourier7_8.zw=pFourierCoefficient[7]*Factor+Summand;
}

/*! Given a scalar this function outputs four vectors, each consisting of two 
   complex Fourier coefficients of the delta distribution realizing this scalar.*/
void ComputeMomentVectorFourier16(out float4 OutFourier1_2,out float4 OutFourier3_4,out float4 OutFourier5_6,out float4 OutFourier7_8,out float4 OutFourier9_10,out float4 OutFourier11_12,out float4 OutFourier13_14,out float4 OutFourier15_16,float FragmentDepth){
	const float PI=3.1415926535897932384626433832795f;
	// Compute the first Fourier coefficient
	float2 Fourier1;
	sincos((FragmentDepth*0.5f+0.5f)*PI,Fourier1.y,Fourier1.x);
	// Compute the other coefficients by complex exponentiation
	float2 FourierSquare=Multiply(Fourier1,Fourier1);
	float2 pFourierCoefficient[16];
	pFourierCoefficient[0]=Fourier1;
	[unroll] for(int i=1;i!=16;++i){
		pFourierCoefficient[i]=Multiply(pFourierCoefficient[i-1],FourierSquare);
	}
	// Store the results
	const float2 Factor=float2(0.5f,0.5f);
	const float2 Summand=Factor;
	OutFourier1_2.xy=pFourierCoefficient[0]*Factor+Summand;
	OutFourier1_2.zw=pFourierCoefficient[1]*Factor+Summand;
	OutFourier3_4.xy=pFourierCoefficient[2]*Factor+Summand;
	OutFourier3_4.zw=pFourierCoefficient[3]*Factor+Summand;
	OutFourier5_6.xy=pFourierCoefficient[4]*Factor+Summand;
	OutFourier5_6.zw=pFourierCoefficient[5]*Factor+Summand;
	OutFourier7_8.xy=pFourierCoefficient[6]*Factor+Summand;
	OutFourier7_8.zw=pFourierCoefficient[7]*Factor+Summand;
	OutFourier9_10.xy=pFourierCoefficient[8]*Factor+Summand;
	OutFourier9_10.zw=pFourierCoefficient[9]*Factor+Summand;
	OutFourier11_12.xy=pFourierCoefficient[10]*Factor+Summand;
	OutFourier11_12.zw=pFourierCoefficient[11]*Factor+Summand;
	OutFourier13_14.xy=pFourierCoefficient[12]*Factor+Summand;
	OutFourier13_14.zw=pFourierCoefficient[13]*Factor+Summand;
	OutFourier15_16.xy=pFourierCoefficient[14]*Factor+Summand;
	OutFourier15_16.zw=pFourierCoefficient[15]*Factor+Summand;
}

/*! Given a scalar this function outputs a vector, consisting of the first two 
   complex Fourier coefficients of the delta distribution realizing this scalar.*/
void ComputeTrigonometricMomentVector(out float4 OutFourier1_2,float FragmentDepth){
	const float PI=3.1415926535897932384626433832795f;
	// Compute the first Fourier coefficient
	float2 Fourier1;
	sincos(FragmentDepth*PI+PI,Fourier1.y,Fourier1.x);
	// Square it to obtain the second Fourier coefficient
	float2 Fourier2=Multiply(Fourier1,Fourier1);
	// Store the results
	const float2 Factor=float2(0.5f,0.5f);
	const float2 Summand=Factor;
	OutFourier1_2.xy=Fourier1*Factor+Summand;
	OutFourier1_2.zw=Fourier2*Factor+Summand;
}

/*! Given a scalar this function sends it through four scaled exponential functions 
   and returns the resulting vector.*/
void ComputeExponentialVarianceVector(out float4 OutDepthExponentialVariance,float FragmentDepth,float2 ExponentialVarianceFactors){
	float SignedDepth=FragmentDepth;
	float Exponential=exp(ExponentialVarianceFactors.x*SignedDepth);
	float InvExponential=exp(ExponentialVarianceFactors.y*SignedDepth);
	OutDepthExponentialVariance=float4(Exponential,Exponential*Exponential,InvExponential,InvExponential*InvExponential);
}

/*! Given a scalar this function sends it through four scaled exponential functions 
   and returns the resulting vector.*/
void ComputeExponentialVarianceVectorSplit(out float2 OutDepthExponentialVariance0,out float2 OutDepthExponentialVariance1,float FragmentDepth,float2 ExponentialVarianceFactors){
	float SignedDepth=FragmentDepth;
	float Exponential=exp(ExponentialVarianceFactors.x*SignedDepth);
	float InvExponential=exp(ExponentialVarianceFactors.y*SignedDepth);
	OutDepthExponentialVariance0=float2(Exponential,Exponential*Exponential);
	OutDepthExponentialVariance1=float2(InvExponential,InvExponential*InvExponential);
}

/*! Given a scalar this function outputs a vector whose first component is the given 
   scalar remapped from [-1,1] to [0,1] and whose second component is the square of 
   the given scalar.*/
void ComputeMomentVector2Moments_float2(out float2 Out2Moments_float2,float FragmentDepth){
	Out2Moments_float2=float2(FragmentDepth*0.5f+0.5f,FragmentDepth*FragmentDepth);
}


/*! Given a scalar this function outputs a vector consisting of the first, second, 
   third and fourth power of this scalar.*/
void ComputeMomentVector4Moments_float4(out float4 Out4Moments_float4,float FragmentDepth){
	float Square=FragmentDepth*FragmentDepth;
	Out4Moments_float4=float4(FragmentDepth,Square,Square*FragmentDepth,Square*Square);
}

/*! Given a scalar this function outputs a vector consisting of the first, second, 
   third and fourth power of this scalar linearly transformed using an optimized 
   basis for optimal quantization.*/
void ComputeMomentVector4MomentsOptimized_float4(out float4 Out4MomentsOptimized_float4,float FragmentDepth){
	float Square=FragmentDepth*FragmentDepth;
	float4 Moments=float4(FragmentDepth,Square,Square*FragmentDepth,Square*Square);
	Out4MomentsOptimized_float4.xz=mul(Moments.xz,float2x2(1.5f,sqrt(3.0f)*0.5f,-2.0f,-sqrt(3.0f)*2.0f/9.0f))+0.5f;
	Out4MomentsOptimized_float4.yw=mul(Moments.yw,float2x2(4.0f,0.5f,-4.0f,0.5f));
}


/*! Given a scalar this function outputs two vectors consisting of the first, 
   second, third and fourth power of this scalar. Splitting it up into two vectors 
   (and two textures) is useful if the alpha channel is needed for alpha blending. 
  \sa ComputeFixedPrecision() */
void ComputeMomentVector4Moments_float2_float2(out float2 Out4Moments0_float2,out float2 Out4Moments1_float2,float FragmentDepth){
	float Square=FragmentDepth*FragmentDepth;
	Out4Moments0_float2=float2(FragmentDepth,Square);
	Out4Moments1_float2=float2(Square*FragmentDepth,Square*Square);
}

/*! Given a scalar this function outputs a vector consisting of the first, second, 
   third and fourth power of this scalar linearly transformed using an optimized 
   basis for optimal quantization. The output is split up into two vectors as with 
   ComputeMomentVector4Moments_float2_float2.
  \sa ComputeFixedPrecision() */
void ComputeMomentVector4MomentsOptimized_float2_float2(out float2 Out4MomentsOptimized0_float2,out float2 Out4MomentsOptimized1_float2,float FragmentDepth){
	float Square=FragmentDepth*FragmentDepth;
	float4 Moments=float4(FragmentDepth,Square,Square*FragmentDepth,Square*Square);
	float4 OptimizedMoments;
	OptimizedMoments.xz=mul(Moments.xz,float2x2(1.5f,sqrt(3.0f)*0.5f,-2.0f,-sqrt(3.0f)*2.0f/9.0f))+0.5f;
	OptimizedMoments.yw=mul(Moments.yw,float2x2(4.0f,0.5f,-4.0f,0.5f));
	Out4MomentsOptimized0_float2=OptimizedMoments.xy;
	Out4MomentsOptimized1_float2=OptimizedMoments.zw;
}


/*! This function computes the transformations for a shadow map camera for a 
   directional light such that the given axis-aligned world-space bounding box is 
   contained in its frustum.
  \note This is a reference implementation only. The function should be implemented 
		on the CPU-side of the application.*/
void ComputeDirectionalLightCamera(out float4x4 OutViewToWorldSpace,out float4x4 OutViewToProjectionSpace,float3 WorldSpaceBoundingBoxMin,float3 WorldSpaceBoundingBoxMax,float3 WorldSpaceLightDirection){
	// Determine the orientation of a view-space coordinate system to use for the 
	// camera
	float3 LeftVectorUnnormalized=cross(float3(0.0f,1.0f,0.0f),WorldSpaceLightDirection);
	float3 LeftVector;
	if(dot(LeftVectorUnnormalized,LeftVectorUnnormalized)>1.0e-8f){
		LeftVector=normalize(LeftVectorUnnormalized);
	}
	else{
		LeftVector=normalize(cross(float3(0.0f,0.0f,1.0f),WorldSpaceLightDirection));
	}
	float3 UpVector=normalize(cross(WorldSpaceLightDirection,LeftVector));
	float3x3 ViewToWorldSpaceRotation=float3x3(
		LeftVector.x,LeftVector.y,LeftVector.z,
		UpVector.x,UpVector.y,UpVector.z,
		WorldSpaceLightDirection.x,WorldSpaceLightDirection.y,WorldSpaceLightDirection.z);
	// Construct all eight vertices of the bounding box
	float3 Min=WorldSpaceBoundingBoxMin;
	float3 Max=WorldSpaceBoundingBoxMax;
	float3 pBoundingBoxVertex[8]={
		float3(Min.x,Min.y,Min.z),
		float3(Min.x,Min.y,Max.z),
		float3(Min.x,Max.y,Min.z),
		float3(Min.x,Max.y,Max.z),
		float3(Max.x,Min.y,Min.z),
		float3(Max.x,Min.y,Max.z),
		float3(Max.x,Max.y,Min.z),
		float3(Max.x,Max.y,Max.z)
	};
	// Compute the minimal and maximal coordinates of these vertices in the camera 
	// coordinate system
	float3 MinExtent=mul(pBoundingBoxVertex[0],transpose(ViewToWorldSpaceRotation));
	float3 MaxExtent=MinExtent;
	[unroll] for(int i=1;i!=8;++i){
		float3 ViewSpaceVertex=mul(pBoundingBoxVertex[i],transpose(ViewToWorldSpaceRotation));
		MinExtent=min(MinExtent,ViewSpaceVertex);
		MaxExtent=max(MaxExtent,ViewSpaceVertex);
	}
	// Compute the camera position
	float3 ViewSpaceCameraPosition=MinExtent*float3(0.5f,0.5f,1.0f)+MaxExtent*float3(0.5f,0.5f,0.0f);
	float3 WorldSpaceCameraPosition=mul(ViewSpaceCameraPosition,ViewToWorldSpaceRotation);
	// Construct the complete view to world space transformation
	OutViewToWorldSpace[0]=float4(ViewToWorldSpaceRotation[0],0.0f);
	OutViewToWorldSpace[1]=float4(ViewToWorldSpaceRotation[1],0.0f);
	OutViewToWorldSpace[2]=float4(ViewToWorldSpaceRotation[2],0.0f);
	OutViewToWorldSpace[3]=float4(WorldSpaceCameraPosition,0.0f);
	// Construct the view to projection space transformation
	float3 ViewSize=MaxExtent-MinExtent;
	OutViewToProjectionSpace=float4x4(
		2.0f/ViewSize.x,0.0f,0.0f,0.0f,
		0.0f,2.0f/ViewSize.y,0.0f,0.0f,
		0.0f,0.0f,1.0f/ViewSize.z,0.0f,
		0.0f,0.0f,0.0f,1.0f);
}


/*! Given various parameters of a spot light this function outputs a view to world 
   space and view to projection space transformation, which can be used for a camera 
   to render a shadow map for this spot light.
  \note This is a reference implementation only. The function should be implemented 
		on the CPU-side of the application.*/
void ComputeSpotLightCamera(out float4x4 OutViewToWorldSpace,out float4x4 OutViewToProjectionSpace,float3 WorldSpaceLightPosition,float3 WorldSpaceConeAxisDirection,float OuterConeAngle,float NearClippingPlane,float FarClippingPlane){
	// Construct a view to world space transformation, which looks into the right 
	// direction
	float3 LeftVector=normalize(cross(float3(0.0f,1.0f,0.0f),WorldSpaceConeAxisDirection));
	float3 UpVector=normalize(cross(WorldSpaceConeAxisDirection,LeftVector));
	OutViewToWorldSpace=float4x4(
		LeftVector.x,LeftVector.y,LeftVector.z,0.0f,
		UpVector.x,UpVector.y,UpVector.z,0.0f,
		WorldSpaceConeAxisDirection.x,WorldSpaceConeAxisDirection.y,WorldSpaceConeAxisDirection.z,0.0f,
		WorldSpaceLightPosition.x,WorldSpaceLightPosition.y,WorldSpaceLightPosition.z,1.0f);
	// Construct the projection transformation using the appropriate field of view
	float Cotangent=1.0/tan(OuterConeAngle);
	float NormalizationFactor=1.0f/(FarClippingPlane-NearClippingPlane);
	OutViewToProjectionSpace=NormalizationFactor*float4x4(
		Cotangent,0.0f,0.0f,0.0f,
		0.0f,Cotangent,0.0f,0.0f,
		0.0f,0.0f,FarClippingPlane*NormalizationFactor,1.0,
		0.0f,0.0f,-NearClippingPlane*FarClippingPlane*NormalizationFactor,0.0f);
}


/*! Given the coordinate of a point in the projection space of the shadow map camera 
   of a light source this function outputs the texture coordinate, which has to be 
   used to look up a shadow map sample for this point and the depth this point would 
   have in the shadow map.
  \note It is non-linear and thus has to be evaluated per pixel.*/
void ComputeShadowMapCoordinate(out float2 OutShadowMapTexCoord,out float OutShadowMapDepth,float4 ShadowMapProjectionSpacePosition,float4 ShadowMapSize){
	OutShadowMapTexCoord=ShadowMapProjectionSpacePosition.xy/ShadowMapProjectionSpacePosition.w;
	OutShadowMapTexCoord=OutShadowMapTexCoord*float2(0.5f,-0.5f)+float2(0.5f,0.5f);
#if D3D_VERSION==9
	OutShadowMapTexCoord=OutShadowMapTexCoord+0.5f*ShadowMapSize.zw;
#endif
	OutShadowMapDepth=(ShadowMapProjectionSpacePosition.z/ShadowMapProjectionSpacePosition.w)*2.0f-1.0f;
}


/*! This function computes the world space position of the surface that is currently 
   visible under the mouse cursor.
  \param OutWorldSpacePosition The world space position of the fragment that last 
		 wrote to the depth buffer at the cursor position. Typically this is the foremost 
		 opaque surface.
  \param SceneDepthTexture The depth stencil texture used during main scene rendering.
  \param SceneDepthTextureSize The size (xy) and reciprocal size (zw) of 
		 SceneDepthTexture in pixels.
  \param iCursorPosition The index of the pixel in the depth stencil texture pointed 
		 to by the mouse cursor.
  \param CameraProjectionToWorldSpace The projection to world space transform of the 
		 camera used for main scene rendering.*/
void ComputeWorldSpacePositionUnderCursor(out float3 OutWorldSpacePosition,Texture2DMS<float,1> SceneDepthTexture,float4 SceneDepthTextureSize,int2 iCursorPosition,float4x4 CameraProjectionToWorldSpace){
	float SceneDepth=SceneDepthTexture.Load(iCursorPosition,0);
	float4 ProjectionSpacePosition;
	ProjectionSpacePosition.x=(float(iCursorPosition.x)+0.5f)*SceneDepthTextureSize.z*2.0f-1.0f;
	ProjectionSpacePosition.y=(float(iCursorPosition.y)+0.5f)*SceneDepthTextureSize.w*(-2.0f)+1.0f;
	ProjectionSpacePosition.z=SceneDepth;
	ProjectionSpacePosition.w=1.0f;
	float4 WorldSpacePosition=mul(ProjectionSpacePosition,CameraProjectionToWorldSpace);
	OutWorldSpacePosition=WorldSpacePosition.xyz/WorldSpacePosition.w;
}


/*! The Heaviside step function. OutShadowIntensity is zero for BlockerDepth>=
   FragmentDepth and one otherwise.*/
void HeavisideStepFunction(out float OutShadowIntensity,float BlockerDepth,float FragmentDepth){
	OutShadowIntensity=(BlockerDepth<FragmentDepth)?1.0f:0.0f;
}


/*! This function takes a sample from the shadow map in the given sampler using the 
   given texture coordinate and outputs the result.*/
void SampleShadowMap(out float OutShadowMapDepth,SSamplerTexturePair2D_float ShadowMapSampler,float2 ShadowMapTexCoord){
	OutShadowMapDepth=ShadowMapSampler.Texture.SampleLevel(ShadowMapSampler.Sampler,ShadowMapTexCoord,0.0f);
}

/*! This function takes a sample from the exponential variance shadow map in the 
   given sampler using the given texture coordinate and outputs it.*/
void SampleExponentialVarianceShadowMap(out float4 OutExponentialVarianceVector,SSamplerTexturePair2D ExponentialVarianceShadowMapSampler0,float2 ShadowMapTexCoord){
#ifdef SHADER_PROFILE_VS
	OutExponentialVarianceVector=ExponentialVarianceShadowMapSampler0.Texture.SampleLevel(ExponentialVarianceShadowMapSampler0.Sampler,ShadowMapTexCoord,0.0f);
#else
	OutExponentialVarianceVector=ExponentialVarianceShadowMapSampler0.Texture.Sample(ExponentialVarianceShadowMapSampler0.Sampler,ShadowMapTexCoord);
#endif
}

/*! This function takes samples from the split exponential variance shadow map in 
   the given two samplers using the given texture coordinate and outputs it the 
   packed result.*/
void SampleExponentialVarianceSplitShadowMap(out float4 OutExponentialVarianceVector,SSamplerTexturePair2D_float2 ExponentialVarianceShadowMapSampler0,SSamplerTexturePair2D_float2 ExponentialVarianceShadowMapSampler1,float2 ShadowMapTexCoord){
#ifdef SHADER_PROFILE_VS
	OutExponentialVarianceVector.xy=ExponentialVarianceShadowMapSampler0.Texture.SampleLevel(ExponentialVarianceShadowMapSampler0.Sampler,ShadowMapTexCoord,0.0f);
	OutExponentialVarianceVector.zw=ExponentialVarianceShadowMapSampler1.Texture.SampleLevel(ExponentialVarianceShadowMapSampler1.Sampler,ShadowMapTexCoord,0.0f);
#else
	OutExponentialVarianceVector.xy=ExponentialVarianceShadowMapSampler0.Texture.Sample(ExponentialVarianceShadowMapSampler0.Sampler,ShadowMapTexCoord);
	OutExponentialVarianceVector.zw=ExponentialVarianceShadowMapSampler1.Texture.Sample(ExponentialVarianceShadowMapSampler1.Sampler,ShadowMapTexCoord);
#endif
}

/*! This function takes a sample from the exponential shadow map in the given 
   sampler using the given texture coordinate and outputs it.*/
void SampleExponentialShadowMap(out float OutExponentialShadowMap,SSamplerTexturePair2D_float ExponentialShadowMapSampler,float2 ShadowMapTexCoord){
#ifdef SHADER_PROFILE_VS
	OutExponentialShadowMap=ExponentialShadowMapSampler.Texture.SampleLevel(ExponentialShadowMapSampler.Sampler,ShadowMapTexCoord,0.0f);
#else
	OutExponentialShadowMap=ExponentialShadowMapSampler.Texture.Sample(ExponentialShadowMapSampler.Sampler,ShadowMapTexCoord);
#endif
}

/*! This function takes a sample from the given set of convolution shadow maps in 
   the given samplers using the given texture coordinate and outputs the complex 
   Fourier coefficients to the given array.*/
void SampleConvolutionShadowMap8(out float2 pOutFourierCoefficient[8],SSamplerTexturePair2D ConvolutionShadowMapSampler0,SSamplerTexturePair2D ConvolutionShadowMapSampler1,SSamplerTexturePair2D ConvolutionShadowMapSampler2,SSamplerTexturePair2D ConvolutionShadowMapSampler3,float2 ShadowMapTexCoord){
	const float2 Factor=float2(2.0f,2.0f);
	const float2 Summand=float2(-1.0f,-1.0f);
#ifdef SHADER_PROFILE_VS
	float4 Fourier1_2=ConvolutionShadowMapSampler0.Texture.SampleLevel(ConvolutionShadowMapSampler0.Sampler,ShadowMapTexCoord,0.0f);
	float4 Fourier3_4=ConvolutionShadowMapSampler1.Texture.SampleLevel(ConvolutionShadowMapSampler1.Sampler,ShadowMapTexCoord,0.0f);
	float4 Fourier5_6=ConvolutionShadowMapSampler2.Texture.SampleLevel(ConvolutionShadowMapSampler2.Sampler,ShadowMapTexCoord,0.0f);
	float4 Fourier7_8=ConvolutionShadowMapSampler3.Texture.SampleLevel(ConvolutionShadowMapSampler3.Sampler,ShadowMapTexCoord,0.0f);
#else
	float4 Fourier1_2=ConvolutionShadowMapSampler0.Texture.Sample(ConvolutionShadowMapSampler0.Sampler,ShadowMapTexCoord);
	float4 Fourier3_4=ConvolutionShadowMapSampler1.Texture.Sample(ConvolutionShadowMapSampler1.Sampler,ShadowMapTexCoord);
	float4 Fourier5_6=ConvolutionShadowMapSampler2.Texture.Sample(ConvolutionShadowMapSampler2.Sampler,ShadowMapTexCoord);
	float4 Fourier7_8=ConvolutionShadowMapSampler3.Texture.Sample(ConvolutionShadowMapSampler3.Sampler,ShadowMapTexCoord);
#endif
	pOutFourierCoefficient[0]=Fourier1_2.xy*Factor+Summand;
	pOutFourierCoefficient[1]=Fourier1_2.zw*Factor+Summand;
	pOutFourierCoefficient[2]=Fourier3_4.xy*Factor+Summand;
	pOutFourierCoefficient[3]=Fourier3_4.zw*Factor+Summand;
	pOutFourierCoefficient[4]=Fourier5_6.xy*Factor+Summand;
	pOutFourierCoefficient[5]=Fourier5_6.zw*Factor+Summand;
	pOutFourierCoefficient[6]=Fourier7_8.xy*Factor+Summand;
	pOutFourierCoefficient[7]=Fourier7_8.zw*Factor+Summand;
}


/*! Like SampleConvolutionShadowMap8 but for sixteen Fourier coefficients.*/
void SampleConvolutionShadowMap16(out float2 pOutFourierCoefficient[16],SSamplerTexturePair2D ConvolutionShadowMapSampler0,SSamplerTexturePair2D ConvolutionShadowMapSampler1,SSamplerTexturePair2D ConvolutionShadowMapSampler2,SSamplerTexturePair2D ConvolutionShadowMapSampler3,SSamplerTexturePair2D ConvolutionShadowMapSampler4,SSamplerTexturePair2D ConvolutionShadowMapSampler5,SSamplerTexturePair2D ConvolutionShadowMapSampler6,SSamplerTexturePair2D ConvolutionShadowMapSampler7,float2 ShadowMapTexCoord){
	const float2 Factor=float2(2.0f,2.0f);
	const float2 Summand=float2(-1.0f,-1.0f);
#ifdef SHADER_PROFILE_VS
	float4 Fourier1_2=ConvolutionShadowMapSampler0.Texture.SampleLevel(ConvolutionShadowMapSampler0.Sampler,ShadowMapTexCoord,0.0f);
	float4 Fourier3_4=ConvolutionShadowMapSampler1.Texture.SampleLevel(ConvolutionShadowMapSampler1.Sampler,ShadowMapTexCoord,0.0f);
	float4 Fourier5_6=ConvolutionShadowMapSampler2.Texture.SampleLevel(ConvolutionShadowMapSampler2.Sampler,ShadowMapTexCoord,0.0f);
	float4 Fourier7_8=ConvolutionShadowMapSampler3.Texture.SampleLevel(ConvolutionShadowMapSampler3.Sampler,ShadowMapTexCoord,0.0f);
	float4 Fourier9_10=ConvolutionShadowMapSampler4.Texture.SampleLevel(ConvolutionShadowMapSampler4.Sampler,ShadowMapTexCoord,0.0f);
	float4 Fourier11_12=ConvolutionShadowMapSampler5.Texture.SampleLevel(ConvolutionShadowMapSampler5.Sampler,ShadowMapTexCoord,0.0f);
	float4 Fourier13_14=ConvolutionShadowMapSampler6.Texture.SampleLevel(ConvolutionShadowMapSampler6.Sampler,ShadowMapTexCoord,0.0f);
	float4 Fourier15_16=ConvolutionShadowMapSampler7.Texture.SampleLevel(ConvolutionShadowMapSampler7.Sampler,ShadowMapTexCoord,0.0f);
#else
	float4 Fourier1_2=ConvolutionShadowMapSampler0.Texture.Sample(ConvolutionShadowMapSampler0.Sampler,ShadowMapTexCoord);
	float4 Fourier3_4=ConvolutionShadowMapSampler1.Texture.Sample(ConvolutionShadowMapSampler1.Sampler,ShadowMapTexCoord);
	float4 Fourier5_6=ConvolutionShadowMapSampler2.Texture.Sample(ConvolutionShadowMapSampler2.Sampler,ShadowMapTexCoord);
	float4 Fourier7_8=ConvolutionShadowMapSampler3.Texture.Sample(ConvolutionShadowMapSampler3.Sampler,ShadowMapTexCoord);
	float4 Fourier9_10=ConvolutionShadowMapSampler4.Texture.Sample(ConvolutionShadowMapSampler4.Sampler,ShadowMapTexCoord);
	float4 Fourier11_12=ConvolutionShadowMapSampler5.Texture.Sample(ConvolutionShadowMapSampler5.Sampler,ShadowMapTexCoord);
	float4 Fourier13_14=ConvolutionShadowMapSampler6.Texture.Sample(ConvolutionShadowMapSampler6.Sampler,ShadowMapTexCoord);
	float4 Fourier15_16=ConvolutionShadowMapSampler7.Texture.Sample(ConvolutionShadowMapSampler7.Sampler,ShadowMapTexCoord);
#endif
	pOutFourierCoefficient[0]=Fourier1_2.xy*Factor+Summand;
	pOutFourierCoefficient[1]=Fourier1_2.zw*Factor+Summand;
	pOutFourierCoefficient[2]=Fourier3_4.xy*Factor+Summand;
	pOutFourierCoefficient[3]=Fourier3_4.zw*Factor+Summand;
	pOutFourierCoefficient[4]=Fourier5_6.xy*Factor+Summand;
	pOutFourierCoefficient[5]=Fourier5_6.zw*Factor+Summand;
	pOutFourierCoefficient[6]=Fourier7_8.xy*Factor+Summand;
	pOutFourierCoefficient[7]=Fourier7_8.zw*Factor+Summand;
	pOutFourierCoefficient[8]=Fourier9_10.xy*Factor+Summand;
	pOutFourierCoefficient[9]=Fourier9_10.zw*Factor+Summand;
	pOutFourierCoefficient[10]=Fourier11_12.xy*Factor+Summand;
	pOutFourierCoefficient[11]=Fourier11_12.zw*Factor+Summand;
	pOutFourierCoefficient[12]=Fourier13_14.xy*Factor+Summand;
	pOutFourierCoefficient[13]=Fourier13_14.zw*Factor+Summand;
	pOutFourierCoefficient[14]=Fourier15_16.xy*Factor+Summand;
	pOutFourierCoefficient[15]=Fourier15_16.zw*Factor+Summand;
}


/*! This function takes a sample from the given shadow map containing two complex 
   Fourier coefficients using the given texture coordinate and outputs the complex 
   Fourier coefficients to the given array.*/
void Sample2TMSMShadowMap(out float2 pOutFourierCoefficient[2],SSamplerTexturePair2D _2TMSMShadowMapSampler,float2 ShadowMapTexCoord){
	const float2 Factor=float2(2.0f,2.0f);
	const float2 Summand=float2(-1.0f,-1.0f);
#ifdef SHADER_PROFILE_VS
	float4 Fourier1_2=_2TMSMShadowMapSampler.Texture.SampleLevel(_2TMSMShadowMapSampler.Sampler,ShadowMapTexCoord,0.0f);
#else
	float4 Fourier1_2=_2TMSMShadowMapSampler.Texture.Sample(_2TMSMShadowMapSampler.Sampler,ShadowMapTexCoord);
#endif
	pOutFourierCoefficient[0]=Fourier1_2.xy*Factor+Summand;
	pOutFourierCoefficient[1]=Fourier1_2.zw*Factor+Summand;
}


/*! This function takes a sample from the two-moment shadow map in the given sampler 
   using the given texture coordinate and outputs it.*/
void Sample2MomentShadowMap(out float2 OutBiased2Moments,SSamplerTexturePair2D_float2 _2MomentShadowMapSampler,float2 ShadowMapTexCoord,float MomentBias=0.0f){
#ifdef SHADER_PROFILE_VS
	OutBiased2Moments=_2MomentShadowMapSampler.Texture.SampleLevel(_2MomentShadowMapSampler.Sampler,ShadowMapTexCoord,0.0f);
#else
	OutBiased2Moments=_2MomentShadowMapSampler.Texture.Sample(_2MomentShadowMapSampler.Sampler,ShadowMapTexCoord);
#endif
	OutBiased2Moments.x=OutBiased2Moments.x*2.0f-1.0f;
	// Guarantee a variance of at least MomentBias
	OutBiased2Moments.y=max(OutBiased2Moments.y,OutBiased2Moments.x*OutBiased2Moments.x+MomentBias);
}

/*! Maps the first given moment from [0,1] to [-1,1] and copies the second moment.*/
void Convert2MomentToCanonical(out float2 OutBiased2Moments,float2 Moments0,float MomentBias=0.0f){
	OutBiased2Moments.x=Moments0.x*2.0f-1.0f;
	// Guarantee a variance of at least MomentBias
	OutBiased2Moments.y=max(Moments0.y,OutBiased2Moments.x*OutBiased2Moments.x+MomentBias);
}


/*! This function takes a sample from the four-moment shadow map in the given 
   sampler using the given texture coordinate converts to biased moments and outputs 
   them.*/
void Sample4MomentShadowMap(out float4 OutBiased4Moments,SSamplerTexturePair2D _4MomentShadowMapSampler,float2 ShadowMapTexCoord,float MomentBias=3.0e-7f){
#ifdef SHADER_PROFILE_VS
	OutBiased4Moments=_4MomentShadowMapSampler.Texture.SampleLevel(_4MomentShadowMapSampler.Sampler,ShadowMapTexCoord,0.0f);
#else
	OutBiased4Moments=_4MomentShadowMapSampler.Texture.Sample(_4MomentShadowMapSampler.Sampler,ShadowMapTexCoord);
#endif
	OutBiased4Moments=lerp(OutBiased4Moments,float4(0.0f,0.375f,0.0f,0.375f),MomentBias);
}

/*! This function takes a sample from the four-moment optimized shadow map in the 
   given sampler using the given texture coordinate converts to biased moments and 
   outputs it.*/
void Sample4MomentOptimizedShadowMap(out float4 OutBiased4Moments,SSamplerTexturePair2D _4MomentOptimizedShadowMapSampler,float2 ShadowMapTexCoord,float MomentBias=6.0e-5f){
#ifdef SHADER_PROFILE_VS
	float4 OptimizedSample=_4MomentOptimizedShadowMapSampler.Texture.SampleLevel(_4MomentOptimizedShadowMapSampler.Sampler,ShadowMapTexCoord,0.0f);
#else
	float4 OptimizedSample=_4MomentOptimizedShadowMapSampler.Texture.Sample(_4MomentOptimizedShadowMapSampler.Sampler,ShadowMapTexCoord);
#endif
	OutBiased4Moments.xz=mul(OptimizedSample.xz-0.5f,float2x2(-1.0f/3.0f,-0.75f,sqrt(3.0f),0.75f*sqrt(3.0f)));
	OutBiased4Moments.yw=mul(OptimizedSample.yw,float2x2(0.125f,-0.125f,1.0f,1.0f));
	OutBiased4Moments=lerp(OutBiased4Moments,float4(0.0f,0.628f,0.0f,0.628f),MomentBias);
}


/*! This function takes a sample from a four-moment shadow map split into two two-
   channel shadow maps using the given texture coordinate and outputs it.*/
void Sample4MomentSplitShadowMap(out float4 OutBiased4Moments,SSamplerTexturePair2D_float2 _4MomentShadowMapSampler0,SSamplerTexturePair2D_float2 _4MomentShadowMapSampler1,float2 ShadowMapTexCoord,float MomentBias=3.0e-7f){
#ifdef SHADER_PROFILE_VS
	OutBiased4Moments.xy=_4MomentShadowMapSampler0.Texture.SampleLevel(_4MomentShadowMapSampler0.Sampler,ShadowMapTexCoord,0.0f);
	OutBiased4Moments.zw=_4MomentShadowMapSampler1.Texture.SampleLevel(_4MomentShadowMapSampler1.Sampler,ShadowMapTexCoord,0.0f);
#else
	OutBiased4Moments.xy=_4MomentShadowMapSampler0.Texture.Sample(_4MomentShadowMapSampler0.Sampler,ShadowMapTexCoord);
	OutBiased4Moments.zw=_4MomentShadowMapSampler1.Texture.Sample(_4MomentShadowMapSampler1.Sampler,ShadowMapTexCoord);
#endif
	OutBiased4Moments=lerp(OutBiased4Moments,float4(0.0f,0.375f,0.0f,0.375f),MomentBias);
}

/*! This function takes a sample from the four-moment optimized shadow map in the 
   given sampler using the given texture coordinate converts to moments and outputs 
   it.*/
void Sample4MomentOptimizedSplitShadowMap(out float4 OutBiased4Moments,SSamplerTexturePair2D_float2 _4MomentOptimizedShadowMapSampler0,SSamplerTexturePair2D_float2 _4MomentOptimizedShadowMapSampler1,float2 ShadowMapTexCoord,float MomentBias=6.0e-5f){
	float4 OptimizedSample;
#ifdef SHADER_PROFILE_VS
	OptimizedSample.xy=_4MomentOptimizedShadowMapSampler0.Texture.SampleLevel(_4MomentOptimizedShadowMapSampler0.Sampler,ShadowMapTexCoord,0.0f);
	OptimizedSample.zw=_4MomentOptimizedShadowMapSampler1.Texture.SampleLevel(_4MomentOptimizedShadowMapSampler1.Sampler,ShadowMapTexCoord,0.0f);
#else
	OptimizedSample.xy=_4MomentOptimizedShadowMapSampler0.Texture.Sample(_4MomentOptimizedShadowMapSampler0.Sampler,ShadowMapTexCoord);
	OptimizedSample.zw=_4MomentOptimizedShadowMapSampler1.Texture.Sample(_4MomentOptimizedShadowMapSampler1.Sampler,ShadowMapTexCoord);
#endif
	OutBiased4Moments.xz=mul(OptimizedSample.xz-0.5f,float2x2(-1.0f/3.0f,-0.75f,sqrt(3.0f),0.75f*sqrt(3.0f)));
	OutBiased4Moments.yw=mul(OptimizedSample.yw,float2x2(0.125f,-0.125f,1.0f,1.0f));
	OutBiased4Moments=lerp(OutBiased4Moments,float4(0.0f,0.628f,0.0f,0.628f),MomentBias);
}


/*! Just applies biasing.*/
void Convert4MomentToCanonical(out float4 OutBiased4Moments,float4 Moments0,float MomentBias=3.0e-7f){
	OutBiased4Moments=Moments0;
	OutBiased4Moments=lerp(OutBiased4Moments,float4(0.0f,0.375f,0.0f,0.375f),MomentBias);
}
/*! Combines the split moment vector into a single vector.*/
void ConvertSplit4MomentToCanonical(out float4 OutBiased4Moments,float2 Moments0,float2 Moments1,float MomentBias=3.0e-7f){
	OutBiased4Moments=float4(Moments0,Moments1);
	OutBiased4Moments=lerp(OutBiased4Moments,float4(0.0f,0.375f,0.0f,0.375f),MomentBias);
}
/*! This function converts a vector of four moments from a representation that is 
   optimized for quantization to the canonical representation.*/
void Convert4MomentOptimizedToCanonical(out float4 OutBiased4Moments,float4 OptimizedMoments0,float MomentBias=6.0e-5f){
	OutBiased4Moments.xz=mul(OptimizedMoments0.xz-0.5f,float2x2(-1.0f/3.0f,-0.75f,sqrt(3.0f),0.75f*sqrt(3.0f)));
	OutBiased4Moments.yw=mul(OptimizedMoments0.yw,float2x2(0.125f,-0.125f,1.0f,1.0f));
	OutBiased4Moments=lerp(OutBiased4Moments,float4(0.0f,0.628f,0.0f,0.628f),MomentBias);
}
/*! Like Convert4MomentOptimizedToCanonical() but takes in a moment vector which is 
   split into two.*/
void ConvertSplit4MomentOptimizedToCanonical(out float4 OutBiased4Moments,float2 OptimizedMoments0,float2 OptimizedMoments1,float MomentBias=6.0e-5f){
	float4 OptimizedMoments=float4(OptimizedMoments0,OptimizedMoments1);
	OutBiased4Moments.xz=mul(OptimizedMoments.xz-0.5f,float2x2(-1.0f/3.0f,-0.75f,sqrt(3.0f),0.75f*sqrt(3.0f)));
	OutBiased4Moments.yw=mul(OptimizedMoments.yw,float2x2(0.125f,-0.125f,1.0f,1.0f));
	OutBiased4Moments=lerp(OutBiased4Moments,float4(0.0f,0.628f,0.0f,0.628f),MomentBias);
}


/*! This function scales the given shadow intensity by a factor greater than 1.0 and 
   clamps the result back into the valid range. In addition it sets the shadow 
   intensity to 0.0, if the given shadow map texture coordinate or depth are outside 
   the valid range.*/
void ScaleShadowIntensity(out float OutShadowIntensity,float ShadowIntensity,float ShadowIntensityFactor,float2 ShadowMapTexCoord,float FragmentDepth){
	OutShadowIntensity=saturate(ShadowIntensity*ShadowIntensityFactor);
	bool OutOfRange=ShadowMapTexCoord.x<0.0f || ShadowMapTexCoord.x>1.0f || 
					ShadowMapTexCoord.y<0.0f || ShadowMapTexCoord.y>1.0f || 
					FragmentDepth<-1.0f || FragmentDepth>1.0f;
	OutShadowIntensity=OutOfRange?0.0f:OutShadowIntensity;
}


/*! Given a sampled value from the shadow map and a computed shadow map depth for a 
   point at the same location this function outputs 1.0, if the fragment is in 
   shadow and 0.0f, if the fragment is lit.*/
void ComputeHardShadowIntensity(out float OutShadowIntensity,float SampledShadowMapDepth,float FragmentDepth,float DepthBias){
	OutShadowIntensity=(SampledShadowMapDepth<mad(FragmentDepth-DepthBias,0.5f,0.5f))?1.0f:0.0f;
}


/*! Performs bilinearly filtered PCF using standard deviation 0.0 and a kernel of 
   size 1*1.
  \sa ComputePCFShadowIntensity() */
void ComputePCFShadowIntensity1x1(out float OutShadowIntensity,SSamplerTexturePair2D_float ShadowMapSampler,float2 ShadowMapTexCoord,float FragmentDepth,float4 ShadowMapSize,float DepthBias){
#if D3D_VERSION==9
	ComputePCFShadowIntensity(OutShadowIntensity,ShadowMapSampler.Texture,ShadowMapSampler.Sampler,ShadowMapTexCoord,FragmentDepth,ShadowMapSize,1.0e-10f,0,DepthBias);
#elif D3D_VERSION==11
	ComputePCFShadowIntensityBilinear(OutShadowIntensity,ShadowMapSampler.Texture,ShadowMapSampler.SamplerComparison,ShadowMapTexCoord,FragmentDepth,ShadowMapSize,1.0e-10f,0,DepthBias);
#endif
}

/*! Performs bilinearly filtered PCF using standard deviation 1.2 and a kernel of 
   size 5*5.
  \sa ComputePCFShadowIntensity() */
void ComputePCFShadowIntensity5x5(out float OutShadowIntensity,SSamplerTexturePair2D_float ShadowMapSampler,float2 ShadowMapTexCoord,float FragmentDepth,float4 ShadowMapSize,float DepthBias){
#if D3D_VERSION==9
	ComputePCFShadowIntensity(OutShadowIntensity,ShadowMapSampler.Texture,ShadowMapSampler.Sampler,ShadowMapTexCoord,FragmentDepth,ShadowMapSize,1.2f,2,DepthBias);
#elif D3D_VERSION==11
	ComputePCFShadowIntensityBilinear(OutShadowIntensity,ShadowMapSampler.Texture,ShadowMapSampler.SamplerComparison,ShadowMapTexCoord,FragmentDepth,ShadowMapSize,1.2f,2,DepthBias);
#endif
}

/*! Performs bilinearly filtered PCF using standard deviation 1.8 and a kernel of 
   size 7*7.
  \sa ComputePCFShadowIntensity() */
void ComputePCFShadowIntensity7x7(out float OutShadowIntensity,SSamplerTexturePair2D_float ShadowMapSampler,float2 ShadowMapTexCoord,float FragmentDepth,float4 ShadowMapSize,float DepthBias){
#if D3D_VERSION==9
	ComputePCFShadowIntensity(OutShadowIntensity,ShadowMapSampler.Texture,ShadowMapSampler.Sampler,ShadowMapTexCoord,FragmentDepth,ShadowMapSize,1.8f,3,DepthBias);
#elif D3D_VERSION==11
	ComputePCFShadowIntensityBilinear(OutShadowIntensity,ShadowMapSampler.Texture,ShadowMapSampler.SamplerComparison,ShadowMapTexCoord,FragmentDepth,ShadowMapSize,1.8f,3,DepthBias);
#endif
}

/*! Performs bilinearly filtered PCF using standard deviation 2.4 and a kernel of 
   size 9*9.
  \sa ComputePCFShadowIntensity() */
void ComputePCFShadowIntensity9x9(out float OutShadowIntensity,SSamplerTexturePair2D_float ShadowMapSampler,float2 ShadowMapTexCoord,float FragmentDepth,float4 ShadowMapSize,float DepthBias){
#if D3D_VERSION==9
	ComputePCFShadowIntensity(OutShadowIntensity,ShadowMapSampler.Texture,ShadowMapSampler.Sampler,ShadowMapTexCoord,FragmentDepth,ShadowMapSize,2.4f,4,DepthBias);
#elif D3D_VERSION==11
	ComputePCFShadowIntensityBilinear(OutShadowIntensity,ShadowMapSampler.Texture,ShadowMapSampler.SamplerComparison,ShadowMapTexCoord,FragmentDepth,ShadowMapSize,2.4f,4,DepthBias);
#endif
}

/*! Performs bilinearly filtered PCF using standard deviation 4.3 and a kernel of 
   size 15*15.
  \sa ComputePCFShadowIntensity() */
void ComputePCFShadowIntensity15x15(out float OutShadowIntensity,SSamplerTexturePair2D_float ShadowMapSampler,float2 ShadowMapTexCoord,float FragmentDepth,float4 ShadowMapSize,float DepthBias){
#if D3D_VERSION==9
	ComputePCFShadowIntensity(OutShadowIntensity,ShadowMapSampler.Texture,ShadowMapSampler.Sampler,ShadowMapTexCoord,FragmentDepth,ShadowMapSize,4.3f,7,DepthBias);
#elif D3D_VERSION==11
	ComputePCFShadowIntensityBilinear(OutShadowIntensity,ShadowMapSampler.Texture,ShadowMapSampler.SamplerComparison,ShadowMapTexCoord,FragmentDepth,ShadowMapSize,4.3f,7,DepthBias);
#endif
}


/*! Performs bilinearly filtered PCF using a hard-coded, randomly rotated Poisson 
   disk sampling.
  \sa ComputePCFShadowIntensity() */
void ComputePCFShadowIntensityPoisson(out float OutShadowIntensity,SSamplerTexturePair2D_float ShadowMapSampler,float2 ShadowMapTexCoord,float FragmentDepth,float4 ShadowMapSize,float DepthBias,float KernelDiameter){
	const float2 pSample[24]={float2(0.222626907535f,0.0376987419418f),float2(-0.288020237452f,0.0483084976027f),float2(0.148792471943f,0.371819322479f),float2(-0.236547109164f,-0.34200277074f),float2(-0.215270313225f,0.3687537531f),float2(0.0504826268158f,-0.510591250885f),float2(0.397889542792f,-0.339257480326f),float2(0.528289416978f,0.193135746913f),float2(-0.624405641543f,-0.00267040033168f),float2(0.042582582866f,0.717798243087f),float2(-0.265949625244f,-0.668857675789f),float2(-0.682615983924f,0.318685080441f),float2(-0.72656536118f,-0.28616597416f),float2(-0.555608194422f,-0.558757632487f),float2(0.813852960936f,-0.0490302757418f),float2(0.0153612941654f,-0.848023558844f),float2(-0.588613523901f,0.611835348797f),float2(0.808496940088f,0.336649119359f),float2(0.536490481955f,0.748683460864f),float2(-0.942823213098f,-0.0715127157808f),float2(-0.164551377044f,0.958722089919f),float2(0.305530222001f,-0.925670811422f),float2(0.815894543795f,-0.539824620967f),float2(0.587381917044f,-0.792860283726f)};
	const float PI=3.1415926535897932384626433832795f;
	// Determine a random rotation of the filter kernel using the shadow map texture 
	// coordinate as seed to achieve temporal coherence.
	float Rotation=2.0f*PI*GetRandomNumber(ShadowMapTexCoord);
	float2x2 RotationMatrix;
	sincos(Rotation,RotationMatrix[0][1],RotationMatrix[0][0]);
	RotationMatrix[1][0]=-RotationMatrix[0][1];
	RotationMatrix[1][1]=RotationMatrix[0][0];
	OutShadowIntensity=0.0f;
	float BiasedDepth=mad(FragmentDepth-DepthBias,0.5f,0.5f);
	[unroll] for(int i=0;i!=24;++i){
		float2 OffsetTexCoord=ShadowMapTexCoord+mul(pSample[i],RotationMatrix)*ShadowMapSize.zw*KernelDiameter;
#if D3D_VERSION==9
		float ShadowIntensity=(ShadowMapSampler.Texture.Sample(ShadowMapSampler.Sampler,OffsetTexCoord)<BiasedDepth)?1.0f:0.0f;
#elif D3D_VERSION==11
		float ShadowIntensity=ShadowMapSampler.Texture.SampleCmp(ShadowMapSampler.SamplerComparison,OffsetTexCoord,BiasedDepth);
#endif
		OutShadowIntensity+=ShadowIntensity/24.0f;
	}
}


/*! Given a sampled value from an exponential shadow map and a computed shadow map 
   depth for a point at the same location this function outputs 1.0, if the fragment 
   is in shadow 0.0f, if the fragment is lit and an intermediate value for partial 
   shadow.*/
void ComputeExponentialShadowIntensity(out float OutShadowIntensity,float ExponentialShadowMapSample,float FragmentDepth,float DepthBias,float Factor=80.0f){
	float BiasedDepth=FragmentDepth-DepthBias;
	OutShadowIntensity=1.0f-exp(-Factor*BiasedDepth)*ExponentialShadowMapSample;
	OutShadowIntensity=saturate(OutShadowIntensity);
}


/*! Given Fourier coefficients from a convolution shadow map and a computed shadow 
   map depth for a point at the same location this function outputs 1.0f, if the 
   fragment is in shadow 0.0f, if the fragment is lit and an intermediate value for 
   partial shadow.*/
void ComputeConvolutionShadowIntensity8(out float OutShadowIntensity,float2 pFourierCoefficient[8],float FragmentDepth,float DepthBias){
	const float PI=3.1415926535897932384626433832795f;
	float BiasedDepth=FragmentDepth-DepthBias;
	// Evaluate the first complex basis function
	float2 Fourier1;
	sincos((BiasedDepth*0.5f+0.5f)*PI,Fourier1.y,Fourier1.x);
	// Evaluate the other basis functions using complex exponentiation
	float2 pBasisFunction[8];
	pBasisFunction[0]=Fourier1;
	float2 FourierSquare=Multiply(Fourier1,Fourier1);
	[unroll] for(int i=1;i!=8;++i){
		pBasisFunction[i]=Multiply(pBasisFunction[i-1],FourierSquare);
	}
	// Compute the shadow intensity which is a linear combination of the basis 
	// functions using the Fourier coefficients as weights
	OutShadowIntensity=0.5f;
	[unroll] for(int i=0;i!=8;++i){
		float Constant=PI*(2.0f*i+1.0f);
		float Summand=pBasisFunction[i].y*pFourierCoefficient[i].x-pBasisFunction[i].x*pFourierCoefficient[i].y;
		OutShadowIntensity+=Summand*2.0f/Constant;
	}
	OutShadowIntensity=saturate(OutShadowIntensity);
}


/*! Like ComputeConvolutionShadowIntensity8() but for 16 Fourier coefficients.*/
void ComputeConvolutionShadowIntensity16(out float OutShadowIntensity,float2 pFourierCoefficient[16],float FragmentDepth,float DepthBias){
	const float PI=3.1415926535897932384626433832795f;
	float BiasedDepth=FragmentDepth-DepthBias;
	// Evaluate the first complex basis function
	float2 Fourier1;
	sincos((BiasedDepth*0.5f+0.5f)*PI,Fourier1.y,Fourier1.x);
	// Evaluate the other basis functions using complex exponentiation
	float2 pBasisFunction[16];
	pBasisFunction[0]=Fourier1;
	float2 FourierSquare=Multiply(Fourier1,Fourier1);
	[unroll] for(int i=1;i!=16;++i){
		pBasisFunction[i]=Multiply(pBasisFunction[i-1],FourierSquare);
	}
	// Compute the shadow intensity which is a linear combination of the basis 
	// functions using the Fourier coefficients as weights
	OutShadowIntensity=0.5f;
	[unroll] for(int i=0;i!=16;++i){
		float Constant=PI*(2.0f*i+1.0f);
		float Summand=pBasisFunction[i].y*pFourierCoefficient[i].x-pBasisFunction[i].x*pFourierCoefficient[i].y;
		OutShadowIntensity+=Summand*2.0f/Constant;
	}
	OutShadowIntensity=saturate(OutShadowIntensity);
}


/*! Given a sampled value from an exponential variance shadow map and a computed 
   shadow map depth for a point at the same location this function outputs 1.0, if 
   the fragment is in shadow 0.0f, if the fragment is lit and an intermediate value 
   for partial shadow.*/
void ComputeExponentialVarianceShadowIntensity(out float OutShadowIntensity,float4 ExponentialVarianceVector,float FragmentDepth,float DepthBias,float2 ExponentialVarianceFactors){
	float BiasedDepth=FragmentDepth-DepthBias;
	// Estimate the shadow intensity using ExponentialVarianceVector.xy
	float b1=ExponentialVarianceVector.x;
	float b2=ExponentialVarianceVector.y;
	float Expectation=b1;
	float Variance=b2-b1*b1;
	float WarpedDepth=exp(ExponentialVarianceFactors.x*BiasedDepth);
	float MinStandardDeviation=1.0e-4f*ExponentialVarianceFactors.x*WarpedDepth;
	Variance=max(Variance,MinStandardDeviation*MinStandardDeviation);
	float MeanDepthDifference=WarpedDepth-Expectation;
	float ShadowIntensity1=(MeanDepthDifference<=0.0f)?0.0f:(1.0f-Variance/(Variance+MeanDepthDifference*MeanDepthDifference));

	// Estimate it using ExponentialVarianceVector.zw
	b1=ExponentialVarianceVector.z;
	b2=ExponentialVarianceVector.w;
	Expectation=b1;
	Variance=b2-b1*b1;
	WarpedDepth=exp(ExponentialVarianceFactors.y*BiasedDepth);
	MinStandardDeviation=1.0e-4f*ExponentialVarianceFactors.y*WarpedDepth;
	Variance=max(Variance,MinStandardDeviation*MinStandardDeviation);
	MeanDepthDifference=WarpedDepth-Expectation;
	float ShadowIntensity2=(MeanDepthDifference>=0.0f)?0.0f:(1.0f-Variance/(Variance+MeanDepthDifference*MeanDepthDifference));
	// Combine the results
	OutShadowIntensity=saturate(max(ShadowIntensity1,ShadowIntensity2));
}


/*! Given a sampled value from a two-moment shadow map (i.e. a variance shadow map) 
   and a computed shadow map depth for a point at the same location this function 
   outputs 1.0, if the fragment is in shadow 0.0f, if the fragment is lit and an 
   intermediate value for partial shadow.*/
void Compute2MomentShadowIntensity(out float OutShadowIntensity,float2 Biased2Moments,float FragmentDepth,float DepthBias){
	float Expectation=Biased2Moments.x;
	float Variance=Biased2Moments.y-Biased2Moments.x*Biased2Moments.x;
	float BiasedDepth=FragmentDepth-DepthBias;
	float MeanDepthDifference=BiasedDepth-Expectation;
	OutShadowIntensity=(MeanDepthDifference<=0.0f)?0.0f:(1.0f-Variance/(Variance+MeanDepthDifference*MeanDepthDifference));
	OutShadowIntensity=saturate(OutShadowIntensity);
}


/*! Given a sampled value from a four-moment shadow map and a computed shadow map 
   depth for a point at the same location this function outputs 1.0, if the fragment 
   is in shadow 0.0f, if the fragment is lit and an intermediate value for partial 
   shadow. The returned value is an optimal lower bound.*/
void Compute4MomentShadowIntensity(out float OutShadowIntensity,
    float4 Biased4Moments,float FragmentDepth,float DepthBias)
{
	// Use short-hands for the many formulae to come
	float4 b=Biased4Moments;
	float3 z;
	z[0]=FragmentDepth-DepthBias;

	// Compute a Cholesky factorization of the Hankel matrix B storing only non-
	// trivial entries or related products
	float L21D11=mad(-b[0],b[1],b[2]);
	float D11=mad(-b[0],b[0], b[1]);
	float SquaredDepthVariance=mad(-b[1],b[1], b[3]);
	float D22D11=dot(float2(SquaredDepthVariance,-L21D11),float2(D11,L21D11));
	float InvD11=1.0f/D11;
	float L21=L21D11*InvD11;

	// Obtain a scaled inverse image of bz=(1,z[0],z[0]*z[0])^T
	float3 c=float3(1.0f,z[0],z[0]*z[0]);
	// Forward substitution to solve L*c1=bz
	c[1]-=b.x;
	c[2]-=b.y+L21*c[1];
	// Scaling to solve D*c2=c1
	c[1]*=InvD11;
	c[2]*=D11/D22D11;
	// Backward substitution to solve L^T*c3=c2
	c[1]-=L21*c[2];
	c[0]-=dot(c.yz,b.xy);
	// Solve the quadratic equation c[0]+c[1]*z+c[2]*z^2 to obtain solutions z[1] 
	// and z[2]
	float InvC2=1.0f/c[2];
	float p=c[1]*InvC2;
	float q=c[0]*InvC2;
	float D=((p*p)/4.0f)-q;
	float r=sqrt(D);
	z[1]=-(p/2.0f)-r;
	z[2]=-(p/2.0f)+r;

	// Use a solution made of four deltas if the solution with three deltas is 
	// invalid
	[branch] if(z[1]<-1.0f || z[2]>1.0f){
		float zFree=((b[0]-b[2])*z[0]+b[3]-b[1])/(z[0]+b[2]-b[0]-b[1]*z[0]);
		float w1Factor=(z[0]>zFree)?1.0f:0.0f;
		// Construct a polynomial taking value zero at z[0] and 1, value 1 at -1 and 
		// value w1Factor at zFree. Start with a linear part and then multiply by 
		// linear factors to get the roots.
		float2 Normalizers;
		Normalizers.x=w1Factor/((zFree-z[0])*mad(zFree,zFree,-1.0f));
		Normalizers.y=0.5f/((zFree+1.0f)*(z[0]+1.0f));
		float4 Polynomial;
		Polynomial[0]=mad(zFree,Normalizers.y,Normalizers.x);
		Polynomial[1]=Normalizers.x-Normalizers.y;
		// Multiply the polynomial by (z-z[0])
		Polynomial[2]=Polynomial[1];
		Polynomial[1]=mad(Polynomial[1],-z[0],Polynomial[0]);
		Polynomial[0]*=-z[0];
		// Multiply the polynomial by (z-1)
		Polynomial[3]=Polynomial[2];
		Polynomial.yz=Polynomial.xy-Polynomial.yz;
		Polynomial[0]*=-1.0f;
		// The shadow intensity is the dot product of the coefficients of this 
		// polynomial and the power moments for the respective powers
		OutShadowIntensity=dot(Polynomial,float4(1.0f,b.xyz));
	}
	// Use the solution with three deltas
	else{
		float4 Switch=
			(z[2]<z[0])?float4(z[1],z[0],1.0f,1.0f):(
			(z[1]<z[0])?float4(z[0],z[1],0.0f,1.0f):
			float4(0.0f,0.0f,0.0f,0.0f));
		float Quotient=(Switch[0]*z[2]-b[0]*(Switch[0]+z[2])+b[1])/((z[2]-Switch[1])*(z[0]-z[1]));
		OutShadowIntensity=Switch[2]+Switch[3]*Quotient;
	}
	OutShadowIntensity=saturate(OutShadowIntensity);
}


/*! Given a sampled value from a four-moment shadow map and a computed shadow map 
   depth for a point at the same location this function outputs 1.0, if the fragment 
   is in shadow 0.0f, if the fragment is lit and an intermediate value for partial 
   shadow. The returned value is an optimal lower bound except for the fact that it 
   does not exploit the knowledge that the original distribution has support in 
   [0,1].*/
void Compute4MomentUnboundedShadowIntensity(out float OutShadowIntensity,
    float4 Biased4Moments,float FragmentDepth,float DepthBias)
{
	// Use short-hands for the many formulae to come
	float4 b=Biased4Moments;
	float3 z;
	z[0]=FragmentDepth-DepthBias;

	// Compute a Cholesky factorization of the Hankel matrix B storing only non-
	// trivial entries or related products
	float L21D11=mad(-b[0],b[1],b[2]);
	float D11=mad(-b[0],b[0], b[1]);
	float SquaredDepthVariance=mad(-b[1],b[1], b[3]);
	float D22D11=dot(float2(SquaredDepthVariance,-L21D11),float2(D11,L21D11));
	float InvD11=1.0f/D11;
	float L21=L21D11*InvD11;
	float D22=D22D11*InvD11;
	float InvD22=1.0f/D22;

	// Obtain a scaled inverse image of bz=(1,z[0],z[0]*z[0])^T
	float3 c=float3(1.0f,z[0],z[0]*z[0]);
	// Forward substitution to solve L*c1=bz
	c[1]-=b.x;
	c[2]-=b.y+L21*c[1];
	// Scaling to solve D*c2=c1
	c[1]*=InvD11;
	c[2]*=InvD22;
	// Backward substitution to solve L^T*c3=c2
	c[1]-=L21*c[2];
	c[0]-=dot(c.yz,b.xy);
	// Solve the quadratic equation c[0]+c[1]*z+c[2]*z^2 to obtain solutions 
	// z[1] and z[2]
	float InvC2=1.0f/c[2];
	float p=c[1]*InvC2;
	float q=c[0]*InvC2;
	float D=(p*p*0.25f)-q;
	float r=sqrt(D);
	z[1]=-p*0.5f-r;
	z[2]=-p*0.5f+r;
	// Compute the shadow intensity by summing the appropriate weights
	float4 Switch=
		(z[2]<z[0])?float4(z[1],z[0],1.0f,1.0f):(
		(z[1]<z[0])?float4(z[0],z[1],0.0f,1.0f):
		float4(0.0f,0.0f,0.0f,0.0f));
	float Quotient=(Switch[0]*z[2]-b[0]*(Switch[0]+z[2])+b[1])/((z[2]-Switch[1])*(z[0]-z[1]));
	OutShadowIntensity=Switch[2]+Switch[3]*Quotient;
	OutShadowIntensity=saturate(OutShadowIntensity);
}


/*! Given a sampled value from a trigonometric moment shadow map providing two 
   complex Fourier coefficients and a computed shadow map depth for a point at the 
   same location this function outputs 1.0, if the fragment is in shadow 0.0f, if 
   the fragment is lit and an intermediate value for partial shadow. The returned 
   value is an optimal lower bound.*/
void Compute2TMSMShadowIntensity(out float OutShadowIntensity,
	float2 pFourierCoefficient[2],float FragmentDepth,float DepthBias,float MomentBias){
	// Apply biasing
	pFourierCoefficient[0]*=(1.0f-MomentBias);
	pFourierCoefficient[1]*=(1.0f-MomentBias);
	float BiasedDepth=FragmentDepth-DepthBias;
	OutShadowIntensity=Get2TMSMMinimalWeight(pFourierCoefficient,BiasedDepth*0.5f+0.5f);
	OutShadowIntensity=saturate(OutShadowIntensity);
}


/*! @{
   The following functions take a depth sample from a multisampled texture.
  \param OutDepthSample The loaded depth sample mapped to the range from -1 to 1.
  \param ShadowMap The multisampled shadow map.
  \param TexCoord The texture coordinate at which sampling should take place. No 
		 interpolation is performed.
  \param SampleIndex The index of the sample in the multisampled texture which is to 
		 be loaded.
  \param ShadowMapSize The size of the texture in texels in x and y.*/
void LoadDepthSample1(out float OutDepthSample,Texture2DMS<float,1> ShadowMap,float2 TexCoord,uint SampleIndex,float4 ShadowMapSize){
	OutDepthSample=ShadowMap.Load(int2(TexCoord*ShadowMapSize.xy),SampleIndex)*2.0f-1.0f;
}
void LoadDepthSample2(out float OutDepthSample,Texture2DMS<float,2> ShadowMap,float2 TexCoord,uint SampleIndex,float4 ShadowMapSize){
	OutDepthSample=ShadowMap.Load(int2(TexCoord*ShadowMapSize.xy),SampleIndex)*2.0f-1.0f;
}
void LoadDepthSample4(out float OutDepthSample,Texture2DMS<float,4> ShadowMap,float2 TexCoord,uint SampleIndex,float4 ShadowMapSize){
	OutDepthSample=ShadowMap.Load(int2(TexCoord*ShadowMapSize.xy),SampleIndex)*2.0f-1.0f;
}
void LoadDepthSample8(out float OutDepthSample,Texture2DMS<float,8> ShadowMap,float2 TexCoord,uint SampleIndex,float4 ShadowMapSize){
	OutDepthSample=ShadowMap.Load(int2(TexCoord*ShadowMapSize.xy),SampleIndex)*2.0f-1.0f;
}
//! @}

/*! @{
   The following functions add two vectors to be stored in a shadow map. They are 
   used to combine moment vectors for different samples of a multisampled depth 
   stencil texture.
  \param OutSum The component-wise sum of the vectors.
  \param LHS,RHS The summands.*/
void AddShadowMapData_float(out float OutSum,float LHS,float RHS){
	OutSum=LHS+RHS;
}
void AddShadowMapData_float2(out float2 OutSum,float2 LHS,float2 RHS){
	OutSum=LHS+RHS;
}
void AddShadowMapData_float4(out float4 OutSum,float4 LHS,float4 RHS){
	OutSum=LHS+RHS;
}
void AddShadowMapData_uint2(out uint2 OutSum,uint2 LHS,uint2 RHS){
	OutSum=LHS+RHS;
}
void AddShadowMapData_uint4(out uint4 OutSum,uint4 LHS,uint4 RHS){
	OutSum=LHS+RHS;
}
void AddShadowMapData_half4(out half4 OutSum,half4 LHS,half4 RHS){
	OutSum=LHS+RHS;
}
//! @}

/*! @{
   The following functions divide a vector of moments by an integer. They are used 
   to compute the average of moment vectors moment vectors for different samples of 
   a multisampled depth stencil texture once the sum has been computed.
  \param OutSampleAverage The result of the component-wise division.
  \param SampleSum The dividend.
  \param nSample The divisor.*/
void ComputeSampleAverage_float(out float OutSampleAverage,float SampleSum,uint nSample){
	OutSampleAverage=SampleSum/float(nSample);
}
void ComputeSampleAverage_float2(out float2 OutSampleAverage,float2 SampleSum,uint nSample){
	OutSampleAverage=SampleSum/float(nSample);
}
void ComputeSampleAverage_float4(out float4 OutSampleAverage,float4 SampleSum,uint nSample){
	OutSampleAverage=SampleSum/float(nSample);
}
void ComputeSampleAverage_uint2(out uint2 OutSampleAverage,uint2 SampleSum,uint nSample){
	OutSampleAverage=SampleSum/nSample;
}
void ComputeSampleAverage_uint4(out uint4 OutSampleAverage,uint4 SampleSum,uint nSample){
	OutSampleAverage=SampleSum/nSample;
}
void ComputeSampleAverage_half4(out half4 OutSampleAverage,half4 SampleSum,uint nSample){
	OutSampleAverage=half4(float4(SampleSum)/float(nSample));
}
//! @}


/*! This function can be used to filter a texture using a horizontal Gaussian with 
   standard deviation 1.2, cut off to obtain a 5-texel kernel.
  \param OutFilteredValue The weighted sum of the values in the kernel.
  \param SourceSampler A sampler for the texture to be filtered, which uses bilinear 
		 interpolation.
  \param TexCoord A grid-aligned texture coordinate, i.e. its x-coordinate is a 
		 multiple of TextureSize.z and its y-coordinate is a multiple of 
		 TextureSize.w.
  \param TextureSize The source texture size in x and y and its reciproque in z and 
		 w.*/
void ApplyGaussianFilterHorizontal5_float4(out float4 OutFilteredValue,SSamplerTexturePair2D SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.343406478631f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-1.2608285375f,0.0f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.2608285375f,0.0f));
}
/*! Vertical Gaussian filter with standard deviation 1.2 and kernel size 5.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical5_float4(out float4 OutFilteredValue,SSamplerTexturePair2D SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.343406478631f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-1.2608285375f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.2608285375f));
}
/*! Horizontal Gaussian filter with standard deviation 1.8 and kernel size 7.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterHorizontal7_float4(out float4 OutFilteredValue,SSamplerTexturePair2D SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.18379524343f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-2.31613202819f,0.0f))+
		0.432713776807f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-0.461496136794f,0.0f))+
		0.325387416687f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.38628316249f,0.0f))+
		0.0581035630769f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(3.0f,0.0f));
}
/*! Vertical Gaussian filter with standard deviation 1.8 and kernel size 7.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical7_float4(out float4 OutFilteredValue,SSamplerTexturePair2D SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.18379524343f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-2.31613202819f))+
		0.432713776807f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-0.461496136794f))+
		0.325387416687f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.38628316249f))+
		0.0581035630769f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,3.0f));
}
/*! Horizontal Gaussian filter with standard deviation 2.4 and kernel size 9.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterHorizontal9_float4(out float4 OutFilteredValue,SSamplerTexturePair2D SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.176635428315f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-3.35259798724f,0.0f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-1.43526128506f,0.0f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.43526128506f,0.0f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(3.35259798724f,0.0f));
}
/*! Vertical Gaussian filter with standard deviation 2.4 and kernel size 9.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical9_float4(out float4 OutFilteredValue,SSamplerTexturePair2D SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.176635428315f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-3.35259798724f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-1.43526128506f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.43526128506f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,3.35259798724f));
}
/*! Horizontal Gaussian filter with standard deviation 4.3 and kernel size 15.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterHorizontal15_float4(out float4 OutFilteredValue,SSamplerTexturePair2D SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.0649299817582f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-6.41300869194f,0.0f))+
		0.116774251219f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-4.43945485337f,0.0f))+
		0.169647595987f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-2.46624934622f,0.0f))+
		0.199094815228f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-0.4932400009f,0.0f))+
		0.188751042112f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.47972988258f,0.0f))+
		0.14455532049f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(3.45281792154f,0.0f))+
		0.0894303463716f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(5.4261789926f,0.0f))+
		0.0268166468338f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(7.0f,0.0f));
}
/*! Vertical Gaussian filter with standard deviation 4.3 and kernel size 15.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical15_float4(out float4 OutFilteredValue,SSamplerTexturePair2D SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.0649299817582f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-6.41300869194f))+
		0.116774251219f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-4.43945485337f))+
		0.169647595987f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-2.46624934622f))+
		0.199094815228f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-0.4932400009f))+
		0.188751042112f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.47972988258f))+
		0.14455532049f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,3.45281792154f))+
		0.0894303463716f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,5.4261789926f))+
		0.0268166468338f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,7.0f));
}


/*! Horizontal Gaussian filter with standard deviation 1.2 and kernel size 5.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterHorizontal5_float2(out float4 OutFilteredValue,SSamplerTexturePair2D_float2 SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue.xy=
		0.343406478631f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-1.2608285375f,0.0f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.2608285375f,0.0f));
	OutFilteredValue.zw=float2(0.0f,1.0f);
}
/*! Vertical Gaussian filter with standard deviation 1.2 and kernel size 5.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical5_float2(out float4 OutFilteredValue,SSamplerTexturePair2D_float2 SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue.xy=
		0.343406478631f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-1.2608285375f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.2608285375f));
	OutFilteredValue.zw=float2(0.0f,1.0f);
}
/*! Horizontal Gaussian filter with standard deviation 1.8 and kernel size 7.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterHorizontal7_float2(out float4 OutFilteredValue,SSamplerTexturePair2D_float2 SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue.xy=
		0.18379524343f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-2.31613202819f,0.0f))+
		0.432713776807f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-0.461496136794f,0.0f))+
		0.325387416687f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.38628316249f,0.0f))+
		0.0581035630769f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(3.0f,0.0f));
	OutFilteredValue.zw=float2(0.0f,1.0f);
}
/*! Vertical Gaussian filter with standard deviation 1.8 and kernel size 7.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical7_float2(out float4 OutFilteredValue,SSamplerTexturePair2D_float2 SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue.xy=
		0.18379524343f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-2.31613202819f))+
		0.432713776807f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-0.461496136794f))+
		0.325387416687f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.38628316249f))+
		0.0581035630769f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,3.0f));
	OutFilteredValue.zw=float2(0.0f,1.0f);
}
/*! Horizontal Gaussian filter with standard deviation 2.4 and kernel size 9.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterHorizontal9_float2(out float4 OutFilteredValue,SSamplerTexturePair2D_float2 SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue.xy=
		0.176635428315f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-3.35259798724f,0.0f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-1.43526128506f,0.0f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.43526128506f,0.0f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(3.35259798724f,0.0f));
	OutFilteredValue.zw=float2(0.0f,1.0f);
}
/*! Vertical Gaussian filter with standard deviation 2.4 and kernel size 9.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical9_float2(out float4 OutFilteredValue,SSamplerTexturePair2D_float2 SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue.xy=
		0.176635428315f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-3.35259798724f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-1.43526128506f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.43526128506f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,3.35259798724f));
	OutFilteredValue.zw=float2(0.0f,1.0f);
}
/*! Horizontal Gaussian filter with standard deviation 4.3 and kernel size 15.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterHorizontal15_float2(out float4 OutFilteredValue,SSamplerTexturePair2D_float2 SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue.xy=
		0.0649299817582f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-6.41300869194f,0.0f))+
		0.116774251219f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-4.43945485337f,0.0f))+
		0.169647595987f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-2.46624934622f,0.0f))+
		0.199094815228f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-0.4932400009f,0.0f))+
		0.188751042112f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.47972988258f,0.0f))+
		0.14455532049f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(3.45281792154f,0.0f))+
		0.0894303463716f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(5.4261789926f,0.0f))+
		0.0268166468338f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(7.0f,0.0f));
	OutFilteredValue.zw=float2(0.0f,1.0f);
}
/*! Vertical Gaussian filter with standard deviation 4.3 and kernel size 15.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical15_float2(out float4 OutFilteredValue,SSamplerTexturePair2D_float2 SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue.xy=
		0.0649299817582f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-6.41300869194f))+
		0.116774251219f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-4.43945485337f))+
		0.169647595987f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-2.46624934622f))+
		0.199094815228f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-0.4932400009f))+
		0.188751042112f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.47972988258f))+
		0.14455532049f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,3.45281792154f))+
		0.0894303463716f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,5.4261789926f))+
		0.0268166468338f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,7.0f));
	OutFilteredValue.zw=float2(0.0f,1.0f);
}


/*! Horizontal Gaussian filter with standard deviation 1.2 and kernel size 5.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterHorizontal5_float(out float4 OutFilteredValue,SSamplerTexturePair2D_float SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.343406478631f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-1.2608285375f,0.0f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.2608285375f,0.0f));
}
/*! Vertical Gaussian filter with standard deviation 1.2 and kernel size 5.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical5_float(out float4 OutFilteredValue,SSamplerTexturePair2D_float SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.343406478631f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-1.2608285375f))+
		0.328296760685f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.2608285375f));
}
/*! Horizontal Gaussian filter with standard deviation 1.8 and kernel size 7.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterHorizontal7_float(out float4 OutFilteredValue,SSamplerTexturePair2D_float SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.18379524343f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-2.31613202819f,0.0f))+
		0.432713776807f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-0.461496136794f,0.0f))+
		0.325387416687f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.38628316249f,0.0f))+
		0.0581035630769f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(3.0f,0.0f));
}
/*! Vertical Gaussian filter with standard deviation 1.8 and kernel size 7.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical7_float(out float4 OutFilteredValue,SSamplerTexturePair2D_float SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.18379524343f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-2.31613202819f))+
		0.432713776807f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-0.461496136794f))+
		0.325387416687f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.38628316249f))+
		0.0581035630769f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,3.0f));
}
/*! Horizontal Gaussian filter with standard deviation 2.4 and kernel size 9.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterHorizontal9_float(out float4 OutFilteredValue,SSamplerTexturePair2D_float SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.176635428315f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-3.35259798724f,0.0f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-1.43526128506f,0.0f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.43526128506f,0.0f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(3.35259798724f,0.0f));
}
/*! Vertical Gaussian filter with standard deviation 2.4 and kernel size 9.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical9_float(out float4 OutFilteredValue,SSamplerTexturePair2D_float SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.176635428315f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,0.0f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-3.35259798724f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-1.43526128506f))+
		0.28676825969f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.43526128506f))+
		0.124914026152f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,3.35259798724f));
}
/*! Horizontal Gaussian filter with standard deviation 4.3 and kernel size 15.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterHorizontal15_float(out float4 OutFilteredValue,SSamplerTexturePair2D_float SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.0649299817582f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-6.41300869194f,0.0f))+
		0.116774251219f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-4.43945485337f,0.0f))+
		0.169647595987f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-2.46624934622f,0.0f))+
		0.199094815228f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(-0.4932400009f,0.0f))+
		0.188751042112f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(1.47972988258f,0.0f))+
		0.14455532049f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(3.45281792154f,0.0f))+
		0.0894303463716f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(5.4261789926f,0.0f))+
		0.0268166468338f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(7.0f,0.0f));
}
/*! Vertical Gaussian filter with standard deviation 4.3 and kernel size 15.
  \sa ApplyGaussianFilterHorizontal5_float4() */
void ApplyGaussianFilterVertical15_float(out float4 OutFilteredValue,SSamplerTexturePair2D_float SourceSampler,float2 TexCoord,float4 TextureSize){
	OutFilteredValue=
		0.0649299817582f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-6.41300869194f))+
		0.116774251219f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-4.43945485337f))+
		0.169647595987f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-2.46624934622f))+
		0.199094815228f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,-0.4932400009f))+
		0.188751042112f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,1.47972988258f))+
		0.14455532049f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,3.45281792154f))+
		0.0894303463716f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,5.4261789926f))+
		0.0268166468338f*SourceSampler.Texture.Sample(SourceSampler.Sampler,TexCoord+float2(TextureSize.z,TextureSize.w)*float2(0.0f,7.0f));
}


/*! Generates horizontal prefix sums of the given source texture and stores them in 
   the given output texture.
  \param OutFilteredTexture The texture which is to be overwritten by prefix sums.
  \param SourceSampler A texture and sampler providing access to the texture for 
		 which prefix sums are to be generated. It must have the same dimensions as 
		 the output texture.
  \param ThreadID The y component of this vector indicates the row on which to 
		 operate.*/
void ApplyPrefixSumHorizontal_float4(inout RWTexture2D<float4> OutFilteredTexture,SSamplerTexturePair2D SourceSampler,uint3 ThreadID){
	uint Width,Height;
	OutFilteredTexture.GetDimensions(Width,Height);
	float4 PrefixSum=float4(0.0f,0.0f,0.0f,0.0f);
	for(uint i=0;i<Width;++i){
		PrefixSum+=mad(SourceSampler.Texture.Load(uint3(i,ThreadID.y,0)),2.0f,-1.0f);
		OutFilteredTexture[uint2(i,ThreadID.y)]=PrefixSum;
	}
}

/*! Like ApplyPrefixSumHorizontal but generates vertical prefix sums for the column 
   indicated by ThreadID.x.*/
void ApplyPrefixSumVertical_float4(inout RWTexture2D<float4> OutFilteredTexture,SSamplerTexturePair2D SourceSampler,uint3 ThreadID){
	uint Width,Height;
	OutFilteredTexture.GetDimensions(Width,Height);
	float4 PrefixSum=float4(0.0f,0.0f,0.0f,0.0f);
	for(uint i=0;i<Height;++i){
		PrefixSum+=SourceSampler.Texture.Load(uint3(ThreadID.x,i,0));
		OutFilteredTexture[uint2(ThreadID.x,i)]=PrefixSum;
	}
}

/*! Like ApplyPrefixSumHorizontal_float4() but operates on four channel textures 
   with 32-bit unsigned integers.
  \sa ComputeFixedPrecision() */
void ApplyPrefixSumHorizontal_uint4(inout RWTexture2D<uint4> OutFilteredTexture,SSamplerTexturePair2D SourceSampler,uint3 ThreadID,float2 FixedPrecision){
	uint Width,Height;
	OutFilteredTexture.GetDimensions(Width,Height);
	uint4 PrefixSum=uint4(0,0,0,0);
	for(uint i=0;i<Width;++i){
		PrefixSum+=uint4(SourceSampler.Texture.Load(uint3(i,ThreadID.y,0))*FixedPrecision.x);
		OutFilteredTexture[uint2(i,ThreadID.y)]=PrefixSum;
	}
}

/*! Like ApplyPrefixSumVertical_float4() but operates on four channel textures with 
   32-bit unsigned integers.*/
void ApplyPrefixSumVertical_uint4(inout RWTexture2D<uint4> OutFilteredTexture,SSamplerTexturePair2D_uint4 SourceSampler,uint3 ThreadID){
	uint Width,Height;
	OutFilteredTexture.GetDimensions(Width,Height);
	uint4 PrefixSum=uint4(0,0,0,0);
	for(uint i=0;i<Height;++i){
		PrefixSum+=SourceSampler.Texture.Load(uint3(ThreadID.x,i,0));
		OutFilteredTexture[uint2(ThreadID.x,i)]=PrefixSum;
	}
}

/*! Like ApplyPrefixSumHorizontal_float4() but operates on two channel textures.*/
void ApplyPrefixSumHorizontal_float2(inout RWTexture2D<float2> OutFilteredTexture,SSamplerTexturePair2D_float2 SourceSampler,uint3 ThreadID){
	uint Width,Height;
	OutFilteredTexture.GetDimensions(Width,Height);
	float2 PrefixSum=float2(0.0f,0.0f);
	for(uint i=0;i<Width;++i){
		PrefixSum+=mad(SourceSampler.Texture.Load(uint3(i,ThreadID.y,0)).xy,2.0f,-1.0f);
		OutFilteredTexture[uint2(i,ThreadID.y)]=PrefixSum;
	}
}

/*! Like ApplyPrefixSumVertical_float4() but operates on two channel textures.*/
void ApplyPrefixSumVertical_float2(inout RWTexture2D<float2> OutFilteredTexture,SSamplerTexturePair2D_float2 SourceSampler,uint3 ThreadID){
	uint Width,Height;
	OutFilteredTexture.GetDimensions(Width,Height);
	float2 PrefixSum=float2(0.0f,0.0f);
	for(uint i=0;i<Height;++i){
		PrefixSum+=SourceSampler.Texture.Load(uint3(ThreadID.x,i,0)).xy;
		OutFilteredTexture[uint2(ThreadID.x,i)]=PrefixSum;
	}
}

/*! Like ApplyPrefixSumHorizontal_float4() but operates on two channel textures with 
   32-bit unsigned integers.*/
void ApplyPrefixSumHorizontal_uint2(inout RWTexture2D<uint2> OutFilteredTexture,SSamplerTexturePair2D_float2 SourceSampler,uint3 ThreadID,float2 FixedPrecision){
	uint Width,Height;
	OutFilteredTexture.GetDimensions(Width,Height);
	uint2 PrefixSum=uint2(0,0);
	for(uint i=0;i<Width;++i){
		PrefixSum+=uint2(SourceSampler.Texture.Load(uint3(i,ThreadID.y,0))*FixedPrecision.x);
		OutFilteredTexture[uint2(i,ThreadID.y)]=PrefixSum;
	}
}

/*! Like ApplyPrefixSumVertical_float4() but operates on two channel textures with 
   32-bit unsigned integers.*/
void ApplyPrefixSumVertical_uint2(inout RWTexture2D<uint2> OutFilteredTexture,SSamplerTexturePair2D_uint2 SourceSampler,uint3 ThreadID){
	uint Width,Height;
	OutFilteredTexture.GetDimensions(Width,Height);
	uint2 PrefixSum=uint2(0,0);
	for(uint i=0;i<Height;++i){
		PrefixSum+=SourceSampler.Texture.Load(uint3(ThreadID.x,i,0));
		OutFilteredTexture[uint2(ThreadID.x,i)]=PrefixSum;
	}
}

/*! Like ApplyPrefixSumHorizontal_float4() but operates on four channel textures 
   with half precision entries.*/
void ApplyPrefixSumHorizontal_half4(inout RWTexture2D<half4> OutFilteredTexture,SSamplerTexturePair2D SourceSampler,uint3 ThreadID){
	uint Width,Height;
	OutFilteredTexture.GetDimensions(Width,Height);
	float4 PrefixSum=float4(0.0f,0.0f,0.0f,0.0f);
	for(uint i=0;i<Width;++i){
		PrefixSum+=SourceSampler.Texture.Load(uint3(i,ThreadID.y,0));
		OutFilteredTexture[uint2(i,ThreadID.y)]=(half4)(PrefixSum);
	}
}

/*! Like ApplyPrefixSumVertical_float4() but operates on four channel textures with 
   half precision entries.*/
void ApplyPrefixSumVertical_half4(inout RWTexture2D<half4> OutFilteredTexture,SSamplerTexturePair2D SourceSampler,uint3 ThreadID){
	uint Width,Height;
	OutFilteredTexture.GetDimensions(Width,Height);
	float4 PrefixSum=float4(0.0f,0.0f,0.0f,0.0f);
	for(uint i=0;i<Height;++i){
		PrefixSum+=SourceSampler.Texture.Load(uint3(ThreadID.x,i,0));
		OutFilteredTexture[uint2(ThreadID.x,i)]=(half4)(PrefixSum);
	}
}

/*! This function computes the number of thread groups to be dispatched for 
   generation of a summed area table.
  \param nOutThreadGroupHorizontal The number of thread groups along each dimension 
		 to be dispatched for the horizontal filter.
  \param nOutThreadGroupVertical The number of thread groups along each dimension 
		 to be dispatched for the vertical filter.
  \param ThreadGroupSizeHorizontal The size of thread groups along each dimension 
		 for the horizontal filter. Only the second entry is supposed to be not one.
  \param ThreadGroupSizeVertical The size of thread groups along each dimension for 
		 the vertical filter. Only the first entry is supposed to be not one.
  \param TextureSize The width and height of the texture for which a summed area 
		 table is to be generated in x and y.*/
void GetSummedAreaTableThreadGroupCount(out uint3 nOutThreadGroupHorizontal,out uint3 nOutThreadGroupVertical,uint3 ThreadGroupSizeHorizontal,uint3 ThreadGroupSizeVertical,float4 TextureSize){
	uint Width=(uint)(TextureSize.x);
	uint Height=(uint)(TextureSize.y);
	nOutThreadGroupHorizontal=uint3(1,Height/ThreadGroupSizeHorizontal.y,1);
	nOutThreadGroupVertical=uint3(Width/ThreadGroupSizeVertical.x,1,1);
}


/*! This function computes the average value (component-wise) of a texture with 
   values from zero to one within a rectangle.
  \param OutAverageValue The average value as values in the range from zero to one.
  \param LeftTop,RightBottom The minimal and maximal texture coordinates which are 
		 still contained in the rectangle. Restrictions to the size of this 
		 rectangle may apply.
  \param SummedAreaTableSampler A texture and sampler for a summed area table which 
		 provides prefix sums of the texture of interest.
  \param TextureSize The width and height of the texture in x and y, reciprocals in 
		 z,w.
  \sa ComputeFixedPrecision() */
void ComputeRectangleAverage_uint4(out float4 OutAverageValue,float2 LeftTop,float2 RightBottom,SSamplerTexturePair2D_uint4 SummedAreaTableSampler,float4 TextureSize,float2 FixedPrecision){
	// The summed area table is essentially off by one because the top left texel 
	// already holds the integral over the top left texel. Compensate for that.
	LeftTop-=TextureSize.zw;
	RightBottom-=TextureSize.zw;
	// Round the texture coordinates to integer texels
	float2 LeftTopPixel=LeftTop*TextureSize.xy;
	float2 RightBottomPixel=RightBottom*TextureSize.xy;
	float2 LeftTopPixelFloor=floor(LeftTopPixel);
	float2 RightBottomPixelFloor=floor(RightBottomPixel);
	int2 iLeftTop=int2(LeftTopPixelFloor);
	int2 iRightBottom=int2(RightBottomPixelFloor);
	float2 LeftTopFactor=float2(1.0f,1.0f)-(LeftTopPixel-LeftTopPixelFloor);
	float2 RightBottomFactor=RightBottomPixel-RightBottomPixelFloor;
	// Sample the summed area table at all relevant locations. The first two indices 
	// determine whether we are at the left top or right bottom of the rectangle, 
	// the latter two determine the pixel offset.
	uint4 Samples[2][2][2][2];
	[unroll] for(int x=0;x!=2;++x){
		int TexelX=(x==0)?iLeftTop.x:iRightBottom.x;
		[unroll] for(int y=0;y!=2;++y){
			int TexelY=(y==0)?iLeftTop.y:iRightBottom.y;
			[unroll] for(int z=0;z!=2;++z){
				[unroll] for(int w=0;w!=2;++w){
					Samples[x][y][z][w]=SummedAreaTableSampler.Texture.Load(int3(TexelX+z,TexelY+w,0));
				}
			}
		}
	}
	// Compute integrals for various rectangles
	float4 pCornerIntegral[2][2];
	[unroll] for(int x=0;x!=2;++x){
		[unroll] for(int y=0;y!=2;++y){
			pCornerIntegral[x][y]=float4(Samples[x][y][0][0]+Samples[x][y][1][1]-Samples[x][y][1][0]-Samples[x][y][0][1]);
		}
	}
	float4 pEdgeIntegral[4]={
		// Right edge
		float4(Samples[1][0][0][1]+Samples[1][1][1][0]-Samples[1][0][1][1]-Samples[1][1][0][0]),
		// Top edge
		float4(Samples[0][0][1][0]+Samples[1][0][0][1]-Samples[0][0][1][1]-Samples[1][0][0][0]),
		// Left edge
		float4(Samples[0][0][0][1]+Samples[0][1][1][0]-Samples[0][0][1][1]-Samples[0][1][0][0]),
		// Bottom edge
		float4(Samples[0][1][1][0]+Samples[1][1][0][1]-Samples[0][1][1][1]-Samples[1][1][0][0])
	};
	float4 CenterIntegral=float4(Samples[0][0][1][1]+Samples[1][1][0][0]-Samples[0][1][1][0]-Samples[1][0][0][1]);
	// Compute the integral over the given rectangle
	float4 Integral=CenterIntegral;
	Integral+=pCornerIntegral[0][0]*(LeftTopFactor.x*LeftTopFactor.y);
	Integral+=pCornerIntegral[0][1]*(LeftTopFactor.x*RightBottomFactor.y);
	Integral+=pCornerIntegral[1][0]*(RightBottomFactor.x*LeftTopFactor.y);
	Integral+=pCornerIntegral[1][1]*(RightBottomFactor.x*RightBottomFactor.y);
	Integral+=pEdgeIntegral[0]*RightBottomFactor.x;
	Integral+=pEdgeIntegral[1]*LeftTopFactor.y;
	Integral+=pEdgeIntegral[2]*LeftTopFactor.x;
	Integral+=pEdgeIntegral[3]*RightBottomFactor.y;
	// Get from a non-normalized integral to moments
	float2 Size=RightBottomPixel-LeftTopPixel;
	OutAverageValue=Integral*FixedPrecision.y/(Size.x*Size.y);
}


/*! Like ComputeRectangleAverage_uint4() but without interpolation. The result 
   refers to a modified version of the given rectangle where corner points have been 
   rounded to the pixel grid. This is substantially faster.
  \sa ComputeFixedPrecision() */
void ComputeIntegerRectangleAverage_uint4(out float4 OutAverageValue,float2 LeftTop,float2 RightBottom,SSamplerTexturePair2D_uint4 SummedAreaTableSampler,float4 TextureSize,float2 FixedPrecision){
	int2 iLeftTop=int2(round(LeftTop*TextureSize.xy))-1;
	int2 iRightBottom=int2(round(RightBottom*TextureSize.xy))-1;
	uint4 Sample[2][2]={
		{SummedAreaTableSampler.Texture.Load(int3(iLeftTop.x,iLeftTop.y,0)),
		 SummedAreaTableSampler.Texture.Load(int3(iLeftTop.x,iRightBottom.y,0))},
		{SummedAreaTableSampler.Texture.Load(int3(iRightBottom.x,iLeftTop.y,0)),
		 SummedAreaTableSampler.Texture.Load(int3(iRightBottom.x,iRightBottom.y,0))}
	};
	uint4 Integral=Sample[0][0]+Sample[1][1]-Sample[0][1]-Sample[1][0];
	uint2 Size=iRightBottom-iLeftTop;
	OutAverageValue=float4(Integral)*FixedPrecision.y/float(Size.x*Size.y);
}


/*! Like ComputeRectangleAverage_uint4 but for summed area-tables providing samples 
   as float4.*/
void ComputeRectangleAverage_float4(out float4 OutAverageValue,float2 LeftTop,float2 RightBottom,SSamplerTexturePair2D SummedAreaTableSampler,float4 TextureSize){
	// The summed area table is essentially off by one because the top left texel 
	// already holds the integral over the top left texel, but then again texture 
	// interpolation places the texel value at 0.5. Compensate for that.
	LeftTop-=0.5f*TextureSize.zw;
	RightBottom-=0.5f*TextureSize.zw;
	// Sample (with interpolation but without mip-mapping) at the given texture 
	// coordinates
	float4 LeftTopSample=SummedAreaTableSampler.Texture.SampleLevel(SummedAreaTableSampler.Sampler,LeftTop,0.0f);
	float4 RightTopSample=SummedAreaTableSampler.Texture.SampleLevel(SummedAreaTableSampler.Sampler,float2(RightBottom.x,LeftTop.y),0.0f);
	float4 LeftBottomSample=SummedAreaTableSampler.Texture.SampleLevel(SummedAreaTableSampler.Sampler,float2(LeftTop.x,RightBottom.y),0.0f);
	float4 RightBottomSample=SummedAreaTableSampler.Texture.SampleLevel(SummedAreaTableSampler.Sampler,RightBottom,0.0f);
	// Compute the difference
	float4 Integral=(LeftTopSample-RightTopSample)+(RightBottomSample-LeftBottomSample);
	float2 Size=(RightBottom-LeftTop)*TextureSize.xy;
	OutAverageValue=mad(Integral/(Size.x*Size.y),0.5f,0.5f);
}


/*! Like ComputeRectangleAverage_uint4 but for two-channel textures.
  \sa ComputeFixedPrecision() */
void ComputeRectangleAverage_uint2(out float2 OutAverageValue,float2 LeftTop,float2 RightBottom,SSamplerTexturePair2D_uint2 SummedAreaTableSampler,float4 TextureSize,float2 FixedPrecision){
	// The summed area table is essentially off by one because the top left texel 
	// already holds the integral over the top left texel. Compensate for that.
	LeftTop-=TextureSize.zw;
	RightBottom-=TextureSize.zw;
	// Round the texture coordinates to integer texels
	float2 LeftTopPixel=LeftTop*TextureSize.xy;
	float2 RightBottomPixel=RightBottom*TextureSize.xy;
	float2 LeftTopPixelFloor=floor(LeftTopPixel);
	float2 RightBottomPixelFloor=floor(RightBottomPixel);
	int2 iLeftTop=int2(LeftTopPixelFloor);
	int2 iRightBottom=int2(RightBottomPixelFloor);
	float2 LeftTopFactor=float2(1.0f,1.0f)-(LeftTopPixel-LeftTopPixelFloor);
	float2 RightBottomFactor=RightBottomPixel-RightBottomPixelFloor;
	// Sample the summed area table at all relevant locations. The first two indices 
	// determine whether we are at the left top or right bottom of the rectangle, 
	// the latter two determine the pixel offset.
	uint2 Samples[2][2][2][2];
	[unroll] for(int x=0;x!=2;++x){
		int TexelX=(x==0)?iLeftTop.x:iRightBottom.x;
		[unroll] for(int y=0;y!=2;++y){
			int TexelY=(y==0)?iLeftTop.y:iRightBottom.y;
			[unroll] for(int z=0;z!=2;++z){
				[unroll] for(int w=0;w!=2;++w){
					Samples[x][y][z][w]=SummedAreaTableSampler.Texture.Load(int3(TexelX+z,TexelY+w,0));
				}
			}
		}
	}
	// Compute integrals for various rectangles
	float2 pCornerIntegral[2][2];
	[unroll] for(int x=0;x!=2;++x){
		[unroll] for(int y=0;y!=2;++y){
			pCornerIntegral[x][y]=float2(Samples[x][y][0][0]+Samples[x][y][1][1]-Samples[x][y][1][0]-Samples[x][y][0][1]);
		}
	}
	float2 pEdgeIntegral[4]={
		// Right edge
		float2(Samples[1][0][0][1]+Samples[1][1][1][0]-Samples[1][0][1][1]-Samples[1][1][0][0]),
		// Top edge
		float2(Samples[0][0][1][0]+Samples[1][0][0][1]-Samples[0][0][1][1]-Samples[1][0][0][0]),
		// Left edge
		float2(Samples[0][0][0][1]+Samples[0][1][1][0]-Samples[0][0][1][1]-Samples[0][1][0][0]),
		// Bottom edge
		float2(Samples[0][1][1][0]+Samples[1][1][0][1]-Samples[0][1][1][1]-Samples[1][1][0][0])
	};
	float2 CenterIntegral=float2(Samples[0][0][1][1]+Samples[1][1][0][0]-Samples[0][1][1][0]-Samples[1][0][0][1]);
	// Compute the integral over the given rectangle
	float2 Integral=CenterIntegral;
	Integral+=pCornerIntegral[0][0]*(LeftTopFactor.x*LeftTopFactor.y);
	Integral+=pCornerIntegral[0][1]*(LeftTopFactor.x*RightBottomFactor.y);
	Integral+=pCornerIntegral[1][0]*(RightBottomFactor.x*LeftTopFactor.y);
	Integral+=pCornerIntegral[1][1]*(RightBottomFactor.x*RightBottomFactor.y);
	Integral+=pEdgeIntegral[0]*RightBottomFactor.x;
	Integral+=pEdgeIntegral[1]*LeftTopFactor.y;
	Integral+=pEdgeIntegral[2]*LeftTopFactor.x;
	Integral+=pEdgeIntegral[3]*RightBottomFactor.y;
	// Get from a non-normalized integral to moments
	float2 Size=RightBottomPixel-LeftTopPixel;
	OutAverageValue=Integral*FixedPrecision.y/(Size.x*Size.y);
}


/*! Like ComputeIntegerRectangleAverage_uint4() but for two-channel textures.*/
void ComputeIntegerRectangleAverage_uint2(out float2 OutAverageValue,float2 LeftTop,float2 RightBottom,SSamplerTexturePair2D_uint2 SummedAreaTableSampler,float4 TextureSize,float2 FixedPrecision){
	int2 iLeftTop=int2(round(LeftTop*TextureSize.xy))-1;
	int2 iRightBottom=int2(round(RightBottom*TextureSize.xy))-1;
	uint2 Sample[2][2]={
		{SummedAreaTableSampler.Texture.Load(int3(iLeftTop.x,iLeftTop.y,0)),
		 SummedAreaTableSampler.Texture.Load(int3(iLeftTop.x,iRightBottom.y,0))},
		{SummedAreaTableSampler.Texture.Load(int3(iRightBottom.x,iLeftTop.y,0)),
		 SummedAreaTableSampler.Texture.Load(int3(iRightBottom.x,iRightBottom.y,0))}
	};
	uint2 Integral=Sample[0][0]+Sample[1][1]-Sample[0][1]-Sample[1][0];
	uint2 Size=iRightBottom-iLeftTop;
	OutAverageValue=float2(Integral)*FixedPrecision.y/float(Size.x*Size.y);
}


/*! Like ComputeRectangleAverage_uint4 but for summed area tables providing samples 
   as float2.*/
void ComputeRectangleAverage_float2(out float2 OutAverageValue,float2 LeftTop,float2 RightBottom,SSamplerTexturePair2D_float2 SummedAreaTableSampler,float4 TextureSize){
	// The summed area table is essentially off by one because the top left texel 
	// already holds the integral over the top left texel, but then again texture 
	// interpolation places the texel value at 0.5. Compensate for that.
	LeftTop-=0.5f*TextureSize.zw;
	RightBottom-=0.5f*TextureSize.zw;
	// Sample (with interpolation but without mip-mapping) at the given texture 
	// coordinates
	float2 LeftTopSample=SummedAreaTableSampler.Texture.SampleLevel(SummedAreaTableSampler.Sampler,LeftTop,0.0f);
	float2 RightTopSample=SummedAreaTableSampler.Texture.SampleLevel(SummedAreaTableSampler.Sampler,float2(RightBottom.x,LeftTop.y),0.0f);
	float2 LeftBottomSample=SummedAreaTableSampler.Texture.SampleLevel(SummedAreaTableSampler.Sampler,float2(LeftTop.x,RightBottom.y),0.0f);
	float2 RightBottomSample=SummedAreaTableSampler.Texture.SampleLevel(SummedAreaTableSampler.Sampler,RightBottom,0.0f);
	// Compute the difference
	float2 Integral=(LeftTopSample-RightTopSample)+(RightBottomSample-LeftBottomSample);
	float2 Size=(RightBottom-LeftTop)*TextureSize.xy;
	OutAverageValue=mad(Integral/(Size.x*Size.y),0.5f,0.5f);
}


/*! This function computes factors that define how much precision is to be used for 
   moments stored in integers. It chooses the precision such that adding up all the 
   integer moments within a kernel of the specified maximal size will not overflow a 
   32-bit unsigned integer.
  \param OutFixedPrecision x holds a factor by which values between zero and one 
		 have to be multiplied to convert them to the unsigned integer 
		 representation with fixed precision. y holds the reciprocal of x.
  \param KernelSizeParameter See EstimatePenumbraSize().
  \param ShadowMapSize The size of the shadow map KernelSizeParameter refers to in x 
		 and y.
  \note This is a reference implementation only and you would typically implement it 
		on the CPU-side of your application.*/
void ComputeFixedPrecision(out float2 OutFixedPrecision,float3x2 KernelSizeParameter,float2 ShadowMapSize){
	float2 MaxKernelSizeInPixels=ceil(1.0f+2.0f*KernelSizeParameter._21_22*ShadowMapSize.xy);
	OutFixedPrecision.x=4294967295.0f/(MaxKernelSizeInPixels.x*MaxKernelSizeInPixels.y);
	OutFixedPrecision.y=1.0f/OutFixedPrecision.x;
}


/*! This function computes the left top and right bottom of the rectangle in which 
   the blocker search is to take place.
  \param OutSearchRegionLeftTop The left top of the search region as texture 
		 coordinate.
  \param OutSearchRegionRightBottom The right bottom of the search region as texture 
		 coordinate.
  \param SearchRegionCenter The center of the search region as texture coordinate.
  \param KernelSizeParameter See KernelSizeParameter in EstimatePenumbraSize().
  \param ShadowMapSize Shadow map resolution in xy, reciprocal in zw.*/
void GetBlockerSearchRectangle(out float2 OutSearchRegionLeftTop,out float2 OutSearchRegionRightBottom,float2 SearchRegionCenter,float3x2 KernelSizeParameter,float4 ShadowMapSize){
	OutSearchRegionLeftTop=saturate(SearchRegionCenter-KernelSizeParameter._21_22-0.5f*ShadowMapSize.zw);
	OutSearchRegionRightBottom=saturate(SearchRegionCenter+KernelSizeParameter._21_22+0.5f*ShadowMapSize.zw);
}


/*! This function computes the left top and right bottom of the rectangle over which 
   the shadow map should be filtered to get contact-hardening shadows.
  \param OutFilterRegionLeftTop The left top of the filter region as texture 
		 coordinate.
  \param OutFilterRegionRightBottom The right bottom of the filter region as texture 
		 coordinate.
  \param FilterRegionCenter The center of the filter region as texture coordinate.
  \param KernelSize See OutKernelSize in EstimatePenumbraSize().*/
void GetShadowFilterRectangle(out float2 OutFilterRegionLeftTop,out float2 OutFilterRegionRightBottom,float2 FilterRegionCenter,float2 KernelSize){
	OutFilterRegionLeftTop=saturate(FilterRegionCenter-KernelSize);
	OutFilterRegionRightBottom=saturate(FilterRegionCenter+KernelSize);
}


/*! This function computes the adequate filter size for generating a contact 
   hardening shadow. It also adapts the depth bias to the filter size.
  \param OutKernelSize The size of the kernel as texture coordinate offset from its 
		 center to the right bottom.
  \param OutDepthBias The depth bias to be used when computing filtered shadows over 
		 the returned kernel. For a maximal kernel size it is MaxDepthBias and it is 
		 proportional to OutKernelSize but may be clamped below.
  \param OccluderDepth The (estimated) depth of the occluding geometry in shadow map 
		 coordintes.
  \param FragmentDepth The depth of the receiving fragment in shadow map 
		 coordinates.
  \param LightParameter Parameters defining the computation of the kernel size 
		 which depend solely on the light source and the shadow map projection. The 
		 first column relates to the horizontal kernel size, the second one to the 
		 vertical kernel size. Row 0 holds the factor to be multiplied onto the 
		 distance between occluder and receiver. Row 1 and 2 hold the gradient and 
		 intercept of the linear function in the denominator which depends on the 
		 occluder depth, respectively.
  \param KernelSizeParameter The first row provides the upper top texture coordinate 
		 offst from the center of the smallest possible filter kernel. The second 
		 row provides the same for the biggest possible kernel. The third row 
		 provides the reciprocal of the horizontal kernel offset (in texture 
		 coordinates) at which the depth bias becomes maximal and the minimal depth 
		 bias as multiple of MaxDepthBias.
  \param MaxDepthBias The depth bias that is to be used for a kernel of maximal 
		 size.*/
void EstimatePenumbraSize(out float2 OutKernelSize,out float OutDepthBias,float OccluderDepth,float FragmentDepth,float3x2 LightParameter,float3x2 KernelSizeParameter,float MaxDepthBias){
	float2 Numerator=LightParameter._11_12*(FragmentDepth-OccluderDepth);
	float2 Denominator=LightParameter._21_22*OccluderDepth+LightParameter._31_32;
	OutKernelSize=max(KernelSizeParameter._11_12,min(KernelSizeParameter._21_22,Numerator/Denominator));
	OutDepthBias=MaxDepthBias*clamp(OutKernelSize.x*KernelSizeParameter._31,KernelSizeParameter._32,1.0f);
}


/*! Computes LightParameter as expected by EstimatePenumbraSize() for a directional 
   light.
  \param OutLightParameter The light parameters as expected by 
		 EstimatePenumbraSize().
  \param ViewToProjectionSpace The projection matrix used for the shadow map of the 
		 directional light.
  \param LightSourceAngle The angle spanned by the directional light across the 
		 hemisphere along the two directions corresponding to rows and columns in 
		 the shadow map.*/
void ComputeDirectionalLightSoftShadowParameters(out float3x2 OutLightParameter,float4x4 ViewToProjectionSpace,float2 LightSourceAngle){
	// Compute the bounding box dimensions of the view frustum of the shadow map
	float3 FrustumExtents=float3(
		2.0f*(ViewToProjectionSpace._11*ViewToProjectionSpace._44-ViewToProjectionSpace._14*ViewToProjectionSpace._41)/
			 (ViewToProjectionSpace._11*ViewToProjectionSpace._11-ViewToProjectionSpace._41*ViewToProjectionSpace._41),
		2.0f*(ViewToProjectionSpace._22*ViewToProjectionSpace._44-ViewToProjectionSpace._24*ViewToProjectionSpace._42)/
			 (ViewToProjectionSpace._22*ViewToProjectionSpace._22-ViewToProjectionSpace._42*ViewToProjectionSpace._42),
		dot(float4(ViewToProjectionSpace._43,-ViewToProjectionSpace._43,-ViewToProjectionSpace._33,ViewToProjectionSpace._33),float4(ViewToProjectionSpace._33,ViewToProjectionSpace._43,ViewToProjectionSpace._34,ViewToProjectionSpace._44))/
			(ViewToProjectionSpace._33*(ViewToProjectionSpace._33-ViewToProjectionSpace._43))
	);
	// Compute a factor that turns a depth difference into a kernel size as texture 
	// coordinate
	OutLightParameter._11=0.5f*tan(0.5f*LightSourceAngle.x)*FrustumExtents.z/FrustumExtents.x;
	OutLightParameter._12=0.5f*tan(0.5f*LightSourceAngle.y)*FrustumExtents.z/FrustumExtents.y;
	// The denominator is constant one
	OutLightParameter._21_22=float2(0.0f,0.0f);
	OutLightParameter._31_32=float2(1.0f,1.0f);
}


/*! This function estimates the average depth of fragments in a filter region which 
   are less than FragmentDepth-DepthBias and outputs it. This is done using two 
   moments of the shadow map depth within this filter region.
  \param OutBlockerSearchShadowIntensity See ComputePCSSAverageBlockerDepth().*/
void Compute2MomentAverageBlockerDepth(out float OutAverageBlockerDepth,out float OutBlockerSearchShadowIntensity,float2 BlockerSearchBiased2Moments,float FragmentDepth,float DepthBias){
	// Estimate how much of the kernel is covered by surfaces closer than the 
	// fragment depth
	float Expectation=BlockerSearchBiased2Moments.x;
	float Variance=BlockerSearchBiased2Moments.y-Expectation*Expectation;
	float BiasedDepth=FragmentDepth-DepthBias;
	float MeanDepthDifference=BiasedDepth-Expectation;
	float BlockerWeight=(MeanDepthDifference<=0.0f)?0.0f:(1.0f-Variance/(Variance+MeanDepthDifference*MeanDepthDifference));
	// Solve for the depth of these surfaces (assuming that there are only two 
	// distinct depths in the kernel)
	OutAverageBlockerDepth=(Expectation-FragmentDepth*(1.0f-BlockerWeight))/BlockerWeight;
	OutAverageBlockerDepth=clamp(OutAverageBlockerDepth,-1.0f,1.0f);
	// Are we in the umbra or fully lit?
	bool Umbra=(BlockerWeight>0.99f);
	OutBlockerSearchShadowIntensity=(Umbra?1.0f:(-1.0f));
}


/*! This function estimates the average depth of fragments in a filter region which 
   are less than FragmentDepth-DepthBias and outputs it. This is done using four 
   moments of the shadow map depth within this filter region.
  \param OutBlockerSearchShadowIntensity See ComputePCSSAverageBlockerDepth().*/
void Compute4MomentAverageBlockerDepth(out float OutAverageBlockerDepth,out float OutBlockerSearchShadowIntensity,float4 BlockerSearchBiased4Moments,float FragmentDepth){
	// Use short-hands for the many formulae to come
	float4 b=BlockerSearchBiased4Moments;
	float3 z;
	z[0]=FragmentDepth;

	// Compute a Cholesky factorization of the Hankel matrix B storing only non-
	// trivial entries or related products
	float L21D11=mad(-b[0],b[1],b[2]);
	float D11=mad(-b[0],b[0], b[1]);
	float SquaredDepthVariance=mad(-b[1],b[1], b[3]);
	float D22D11=dot(float2(SquaredDepthVariance,-L21D11),float2(D11,L21D11));
	float InvD11=1.0f/D11;
	float L21=L21D11*InvD11;

	// Obtain a scaled inverse image of bz=(1,z[0],z[0]*z[0])^T
	float3 c=float3(1.0f,z[0],z[0]*z[0]);
	// Forward substitution to solve L*c1=bz
	c[1]-=b.x;
	c[2]-=b.y+L21*c[1];
	// Scaling to solve D*c2=c1
	c[1]*=InvD11;
	c[2]*=D11/D22D11;
	// Backward substitution to solve L^T*c3=c2
	c[1]-=L21*c[2];
	c[0]-=dot(c.yz,b.xy);
	// Solve the quadratic equation c[0]+c[1]*z+c[2]*z^2 to obtain solutions 
	// z[1] and z[2]
	z.yz=GetRoots(c);
	// Compute weights of the Dirac-deltas at the roots
	float3 Weight;
	Weight[0]=(z[1]*z[2]-b[0]*(z[1]+z[2])+b[1])/((z[0]-z[1])*(z[0]-z[2]));
	Weight[1]=(z[0]*z[2]-b[0]*(z[0]+z[2])+b[1])/((z[2]-z[1])*(z[0]-z[1]));
	Weight[2]=1.0f-Weight[0]-Weight[1];
	// Compute the shadow intensity and the unnormalized average depth of occluders
	float AverageBlockerDepthIntegral=((z[1]<z[0])?(Weight[1]*z[1]):0.0f)+((z[2]<z[0])?(Weight[2]*z[2]):0.0f);
	float BlockerSearchShadowIntensity=((z[1]<z[0])?Weight[1]:0.0f)+((z[2]<z[0])?Weight[2]:0.0f);
	const float FullyLitBias=1.0e-3f;
	OutAverageBlockerDepth=clamp((FullyLitBias*z[0]+AverageBlockerDepthIntegral)/(FullyLitBias+BlockerSearchShadowIntensity),-1.0f,1.0f);
	OutBlockerSearchShadowIntensity=(BlockerSearchShadowIntensity>0.99f)?1.0f:(-1.0f);
}


/*! This function computes the average depth of fragments in a filter region which 
   are less than FragmentDepth-DepthBias and outputs it. Samples are taken from the 
   given shadow map in filter kernel whose size corresponds to the maximal admitted 
   size according to the shadow map parameters. The kernel is centered at 
   ShadowMapTexCoord. If all samples are less than FragmentDepth-DepthBias, it 
   outputs FragmentDepth to avoid that time is wasted on filtering a large kernel.
  \param OutBlockerSearchShadowIntensity If the information obtained during the 
		 blocker search admits a definite statement on the appropriate shadow 
		 intensity, this value is set accordingly. Otherwise it is set to -1.0f.*/
void ComputePCSSAverageBlockerDepth(out float OutAverageBlockerDepth,out float OutBlockerSearchShadowIntensity,SSamplerTexturePair2D_float ShadowMapSampler,float2 ShadowMapTexCoord,float FragmentDepth,float4 ShadowMapSize,float DepthBias,float3x2 KernelSizeParameter){
	OutAverageBlockerDepth=0.0f;
	float TotalWeight=0.0f;
	float2 KernelSizePixels=KernelSizeParameter._21_22*ShadowMapSize.xy;
	float2 iShadowMapTexel=ShadowMapTexCoord*ShadowMapSize.xy;
	int2 iLeftTop=int2(iShadowMapTexel-KernelSizePixels);
	int2 iRightBottom=iLeftTop+int2(2.0001f*KernelSizePixels);
	DepthBias*=KernelSizeParameter._32;
	float BiasedDepth=mad(FragmentDepth-DepthBias,0.5f,0.5f);
	for(int x=iLeftTop.x;x<=iRightBottom.x;++x){
		for(int y=iLeftTop.y;y<=iRightBottom.y;++y){
			float Depth=ShadowMapSampler.Texture.Load(int3(x,y,0));
			float Weight=(Depth<BiasedDepth)?1.0f:0.0f;
			OutAverageBlockerDepth+=Weight*Depth;
			TotalWeight+=Weight;
		}
	}
	OutAverageBlockerDepth=(TotalWeight<=0.0f)?1.0f:OutAverageBlockerDepth/TotalWeight;
	OutAverageBlockerDepth=OutAverageBlockerDepth*2.0f-1.0f;
	// Are we in the umbra or fully lit?
	float2 KernelSize=1.0f+float2(iRightBottom-iLeftTop);
	float nKernelTexel=KernelSize.x*KernelSize.y;
	bool Umbra=(TotalWeight>nKernelTexel*0.99f);
	bool FullyLit=(TotalWeight<nKernelTexel*0.01f);
	OutBlockerSearchShadowIntensity=(Umbra?1.0f:(FullyLit?0.0f:(-1.0f)));
}


/*! This function takes in a shadow intensity computed by the filtering step of an 
   algorithm for contact-hardening shadows and another shadow intensity computed by 
   the blocker search. The shadow intensity computed by the blocker search can be 
   negative to imply that results are inconclusive. Only in this case the shadow 
   intensity computed during filtering will be used. Otherwise the shadow intensity 
   computed during the blocker search is output and the compiler can branch away the 
   entire filtering step.
  \warning This method is heavily reliant on compiler optimizations so the produced 
		   assembly should be checked to see whether branches have been created as 
		   intended.*/
void PickShadowIntensity(out float OutShadowIntensity,float FilteredShadowIntensity,float BlockerSearchShadowIntensity){
	[branch] if(BlockerSearchShadowIntensity>=0.0f){
		OutShadowIntensity=BlockerSearchShadowIntensity;
	}
	else{
		OutShadowIntensity=FilteredShadowIntensity;
	}
}


/*! This function evaluates percentage closer filtering for a rectangular kernel 
   with the given dynamically computed corners (in texture coordinates). Texture 
   coordinates are rounded to the texel grid such that texels contribute either 
   entirely or not at all. In this sense there is no interpoltion. SampleCmp() is 
   used to quarter the number of required samples.*/
void ComputeIntegerPCSSShadowIntensity(out float OutShadowIntensity,SSamplerTexturePair2D_float ShadowMapSampler,float2 FilterRegionLeftTop,float2 FilterRegionRightBottom,float FragmentDepth,float4 ShadowMapSize,float DepthBias){
	float BiasedDepth=mad(FragmentDepth-DepthBias,0.5f,0.5f);
	// Round the corners of the rectangle to the grid
	float2 iLeftTop=round(FilterRegionLeftTop*ShadowMapSize.xy);
	float2 iRightBottom=round(FilterRegionRightBottom*ShadowMapSize.xy);
	// Determine the number of required samples and whether the tap count is odd
	float2 nFractionalSample=0.5f*(iRightBottom-iLeftTop)-0.01f;
	float2 nSample=ceil(nFractionalSample);
	bool2 TapCountOdd=(nFractionalSample+0.25f<nSample);
	// Prepare sampling of the last row and last column
	float2 RightBottomWeight=TapCountOdd?float2(1.0f,1.0f):float2(2.0f,2.0f);
	float2 RightBottomTexCoord=(TapCountOdd?(iRightBottom-0.5f):(iRightBottom-1.0f))*ShadowMapSize.zw;
	// Sample everything except for the right column
	float2 InteriorLeftTop=(iLeftTop+1.0f)*ShadowMapSize.zw;
	float2 StepSize=2.0f*ShadowMapSize.zw;
	float2 InteriorRightBottom=InteriorLeftTop+(nSample-1.0f)*StepSize-ShadowMapSize.zw;
	OutShadowIntensity=0.0f;
	float TotalWeight=0.0f;
	[loop] for(float x=InteriorLeftTop.x;x<InteriorRightBottom.x;x+=StepSize.x){
		[loop] for(float y=InteriorLeftTop.y;y<InteriorRightBottom.y;y+=StepSize.y){
			OutShadowIntensity+=ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,float2(x,y),BiasedDepth);
			TotalWeight+=1.0f;
		}
		// Here is a good spot to take care of the bottom row
		OutShadowIntensity+=0.5f*RightBottomWeight.y*ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,float2(x,RightBottomTexCoord.y),BiasedDepth);
		TotalWeight+=0.5f*RightBottomWeight.y;
	}
	OutShadowIntensity*=2.0f;
	TotalWeight*=2.0f;
	// Now take care of the right column
	[loop] for(float y=InteriorLeftTop.y;y<InteriorRightBottom.y;y+=StepSize.y){
		OutShadowIntensity+=RightBottomWeight.x*ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,float2(RightBottomTexCoord.x,y),BiasedDepth);
		TotalWeight+=RightBottomWeight.x;
	}
	OutShadowIntensity*=2.0f;
	TotalWeight*=2.0f;
	// Finish it with the bottom right
	OutShadowIntensity+=RightBottomWeight.x*RightBottomWeight.y*ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,RightBottomTexCoord,BiasedDepth);
	TotalWeight+=RightBottomWeight.x*RightBottomWeight.y;
	// Normalize
	float2 KernelSize=iRightBottom-iLeftTop;
	OutShadowIntensity/=KernelSize.x*KernelSize.y;
}


/*! This function evaluates percentage closer filtering for a rectangular kernel 
   with the given dynamically computed corners (in texture coordinates). Texels are 
   weighted according to the area covered by the specified rectangle. This results 
   in bilinear interpolation. SampleCmp() is used to quarter the number of required 
   samples.*/
void ComputePCSSShadowIntensity(out float OutShadowIntensity,float BlockerSearchShadowIntensity,SSamplerTexturePair2D_float ShadowMapSampler,float2 FilterRegionLeftTop,float2 FilterRegionRightBottom,float FragmentDepth,float4 ShadowMapSize,float DepthBias){
	// Abort early if the blocker search solved the problem already
	[branch] if(BlockerSearchShadowIntensity>=0.0f){
		OutShadowIntensity=BlockerSearchShadowIntensity;
		return;
	}
	float BiasedDepth=mad(FragmentDepth-DepthBias,0.5f,0.5f);
	// Figure out the geometry of the kernel with respect to the pixel grid
	float2 iLeftTop;
	float2 LeftTopFactor=modf(FilterRegionLeftTop*ShadowMapSize.xy,iLeftTop);
	float2 iRightBottom;
	float2 RightBottomFactor=modf(FilterRegionRightBottom*ShadowMapSize.xy,iRightBottom);
	// Determine how many samples will be needed (one sample fetches four adjacent 
	// texels)
	float2 nFractionalSample=0.5f*(iRightBottom+float2(0.99f,0.99f)-iLeftTop);
	float2 nSample=ceil(nFractionalSample);
	bool2 TapCountOdd=(nFractionalSample+0.25f<nSample);
	// Determine where the first and last samples need to be located between texels 
	// to get the right relative weighting and how much weight they need in total
	float2 LeftTopWeight=2.0f-LeftTopFactor;
	float2 LeftTopOffset=1.0f/LeftTopWeight;
	float2 RightBottomWeight=TapCountOdd?RightBottomFactor:(1.0f+RightBottomFactor);
	float2 RightBottomOffset=TapCountOdd?float2(0.0f,0.0f):(1.0f-1.0f/RightBottomWeight);
	// Handle the case of kernels with a single column and/or a single row of 
	// samples
	bool2 SingleSample=(nSample<1.5f);
	float2 SingleWeight=(1.0f-LeftTopFactor)+RightBottomFactor;
	float2 SingleOffset=RightBottomFactor/SingleWeight;
	LeftTopWeight=SingleSample?SingleWeight:LeftTopWeight;
	LeftTopOffset=SingleSample?SingleOffset:LeftTopOffset;
	RightBottomWeight=SingleSample?float2(0.0f,0.0f):RightBottomWeight;
	RightBottomOffset=SingleSample?float2(0.0f,0.0f):RightBottomOffset;
	// Sample the interior of the filter kernel (where weighting is trivial). While 
	// taking care of this it is opportune to also take care of the top and bottom 
	// edge.
	float TotalOcclusion=0.0f;
	float2 LeftTopTexCoord=(iLeftTop+0.5f+LeftTopOffset)*ShadowMapSize.zw;
	float2 RightBottomTexCoord=(iLeftTop-1.5f+2.0f*nSample+RightBottomOffset)*ShadowMapSize.zw;
	float2 StepSize=2.0f*ShadowMapSize.zw;
	float2 InteriorLeftTop=(iLeftTop+3.0f)*ShadowMapSize.zw;
	float2 InteriorPastEnd=InteriorLeftTop+(nSample-2.5f)*StepSize;
	[loop] for(float x=InteriorLeftTop.x;x<InteriorPastEnd.x;x+=StepSize.x){
		// Top
		TotalOcclusion+=0.5f*LeftTopWeight.y*ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,float2(x,LeftTopTexCoord.y),BiasedDepth);
		// Interior
		[loop] for(float y=InteriorLeftTop.y;y<InteriorPastEnd.y;y+=StepSize.y){
			TotalOcclusion+=ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,float2(x,y),BiasedDepth);
		}
		// Bottom
		TotalOcclusion+=0.5f*RightBottomWeight.y*ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,float2(x,RightBottomTexCoord.y),BiasedDepth);
	}
	TotalOcclusion*=4.0f;
	// Sample the remainder of the outer rim piece by piece
	// Left top
	TotalOcclusion+=LeftTopWeight.x*LeftTopWeight.y*ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,LeftTopTexCoord,BiasedDepth);
	// Right top
	if(!SingleSample.x){
		TotalOcclusion+=RightBottomWeight.x*LeftTopWeight.y*ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,float2(RightBottomTexCoord.x,LeftTopTexCoord.y),BiasedDepth);
	}
	// Continue with the vertical edges
	[loop] for(float y=InteriorLeftTop.y;y<InteriorPastEnd.y;y+=StepSize.y){
		// Left
		TotalOcclusion+=LeftTopWeight.x*2.0f*ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,float2(LeftTopTexCoord.x,y),BiasedDepth);
		// Right
		TotalOcclusion+=RightBottomWeight.x*2.0f*ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,float2(RightBottomTexCoord.x,y),BiasedDepth);
	}
	// Left bottom
	if(!SingleSample.y){
		TotalOcclusion+=LeftTopWeight.x*RightBottomWeight.y*ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,float2(LeftTopTexCoord.x,RightBottomTexCoord.y),BiasedDepth);
	}
	// Right bottom
	if(!SingleSample.x && !SingleSample.y){
		TotalOcclusion+=RightBottomWeight.x*RightBottomWeight.y*ShadowMapSampler.Texture.SampleCmpLevelZero(ShadowMapSampler.SamplerComparison,RightBottomTexCoord,BiasedDepth);
	}
	// Divide by the area of the filter kernel to get the end result
	float2 KernelSize=(FilterRegionRightBottom-FilterRegionLeftTop)*ShadowMapSize.xy;
	TotalOcclusion/=KernelSize.x*KernelSize.y;
	OutShadowIntensity=TotalOcclusion;
}
